import judo
from judo.functions.random import random_state
import plangym
import pytest

from fragile.algorithms.fmc import FMCSwarm
from fragile.callbacks.root_walker import BestWalker
from fragile.callbacks.time_steps import GaussianDt
from fragile.callbacks.tree import HistoryTree
from fragile.core.env import PlangymEnv
from fragile.core.policy import Discrete, Gaussian, RandomPlangym
from fragile.core.swarm import Swarm
from fragile.core.walkers import Walkers


if judo.Backend.is_torch():
    pytest.skip(allow_module_level=True)


def create_pacman():
    plangym_env = plangym.make(name="ALE/MsPacman-v5", obs_type="ram", frameskip=5)
    walkers = Walkers(score_scale=2, accumulate_reward=True)
    env = PlangymEnv(plangym_env)
    callbacks = [BestWalker(), GaussianDt(low=2, high=15, loc=5, scale=5)]
    n_walkers = 40
    swarm = Swarm(
        policy=Discrete(),
        walkers=walkers,
        env=env,
        callbacks=callbacks,
        n_walkers=n_walkers,
        max_epochs=100,
    )
    return swarm


def create_cartpole(max_epochs=140):
    plangym_env = plangym.make(name="CartPole-v0")
    walkers = Walkers(score_scale=2, accumulate_reward=True, freeze_walkers=False)
    env = PlangymEnv(plangym_env)
    callbacks = [BestWalker(), HistoryTree(prune=True)]
    n_walkers = 64
    swarm = Swarm(
        policy=RandomPlangym(),
        walkers=walkers,
        env=env,
        callbacks=callbacks,
        n_walkers=n_walkers,
        max_epochs=max_epochs,
    )
    return swarm


def create_dm_cartpole():
    plangym_env = plangym.make(name="cartpole-balance", frameskip=2)
    walkers = Walkers(score_scale=2, accumulate_reward=True)
    env = PlangymEnv(plangym_env)
    policy = Gaussian(scale=2, second_order=True)
    callbacks = [BestWalker(), GaussianDt(low=2, high=15, loc=5, scale=5)]
    n_walkers = 32
    swarm = Swarm(
        policy=policy,
        walkers=walkers,
        env=env,
        callbacks=callbacks,
        n_walkers=n_walkers,
        max_epochs=150,
    )
    return swarm


def cartpole_fmc():
    swarm = create_cartpole(max_epochs=10)
    fmc = FMCSwarm(
        swarm=swarm,
        max_epochs=50,
        callbacks=[BestWalker(always_update=True), HistoryTree(names=["states", "actions"])],
    )
    return fmc


swarms = [
    (create_pacman(), 1800),
    (create_cartpole(), 135),
    # (create_dm_cartpole(), 50),
    (cartpole_fmc(), 48),
]


@pytest.mark.parametrize("swarm_and_target", swarms)
def test_score_gets_higher(swarm_and_target):
    if swarm_and_target is None:
        return
    swarm, target_score = swarm_and_target
    random_state.seed(160290)
    swarm.run()
    root_score = float(swarm.root.score)
    beats_target = root_score <= target_score if swarm.minimize else root_score >= target_score
    assert beats_target, (root_score, target_score)



import warnings


warnings.filterwarnings(
    "ignore",
    message=(
        "Using or importing the ABCs from 'collections' instead of from 'collections.abc' "
        "is deprecated since Python 3.3,and in 3.9 it will stop working"
    ),
)
warnings.filterwarnings(
    "ignore",
    message=(
        "the imp module is deprecated in favour of importlib; see the module's "
        "documentation for alternative uses"
    ),
)
warnings.filterwarnings(
    "ignore",
    message=(
        "Using or importing the ABCs from 'collections' instead of from "
        "'collections.abc' is deprecated, and in 3.8 it will stop working"
    ),
)
warnings.filterwarnings(
    "ignore",
    message=(
        "The set_clim function was deprecated in Matplotlib 3.1 "
        "and will be removed in 3.3. Use ScalarMappable.set_clim "
        "instead."
    ),
)
warnings.filterwarnings(
    "ignore",
    message=(
        "The datapath rcparam was deprecated in Matplotlib 3.2.1 and will be "
        "removed two minor releases later."
    ),
)
warnings.filterwarnings("ignore", message="Gdk.Cursor.new is deprecated")



from typing import Optional, Union

from judo import random_state
import numpy

from fragile.core.policy import Gaussian
from fragile.core.typing import StateData, StateDict, Tensor


class ParticleIntegration(Gaussian):
    default_outputs = ("actions", "velocities")
    default_inputs = {"velocities": {"clone": True}, "dt": {"optional": True}}

    @property
    def param_dict(self) -> StateDict:
        params = super(ParticleIntegration, self).param_dict
        if self.swarm is None:
            return params
        return {**{"velocities": self.swarm.env.param_dict["observs"]}, **params}

    def act(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        """Calculate SwarmState containing the data needed to interact with the environment."""
        action_input = self._prepare_tensors(**kwargs)
        actions_data = self.select_actions(**action_input)
        if not isinstance(actions_data, dict):
            actions_data = {"actions": actions_data}
        actions = actions_data["actions"]
        actions_data["actions"] = self.env_bounds.clip(actions)
        if inplace:
            self.update(**actions_data)
        else:
            return actions_data

    def select_actions(self, **kwargs) -> Union[Tensor, StateData]:
        actions = super(ParticleIntegration, self).select_actions(**kwargs)
        dt = self.get("dt", default=1.0, raise_error=False)
        velocity_0 = self.get("velocities")
        velocities = velocity_0 + actions * dt
        actions = velocity_0 + velocities * dt
        return {"velocities": velocities, "actions": actions}

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ) -> Union[None, StateData]:
        data = self.select_actions(**kwargs)
        if inplace:
            self.update(velocities=data["velocities"])
        return data


class ESModel(Gaussian):
    """
    The ESModel implements an evolutionary strategy policy.

    It mutates randomly some of the coordinates of the best solution found by \
    substituting them with a proposal solution. This proposal solution is the \
    difference between two random permutations of the best solution found.

    It applies a gaussian normal perturbation with a probability given by ``mutation``.
    """

    def __init__(
        self,
        mutation: float = 0.5,
        recombination: float = 0.7,
        random_step_prob: float = 0.1,
        *args,
        **kwargs,
    ):
        """
        Initialize a :class:`ESModel`.

        Args:
            mutation: Probability of mutating a coordinate of the solution vector.
            recombination: Step size of the update applied to the best solution found.
            random_step_prob: Probability of applying a random normal perturbation.
            *args: Passed to the parent :class:`NormalContinuous`.
            **kwargs: Passed to the parent :class:`NormalContinuous`.

        """
        super(ESModel, self).__init__(*args, **kwargs)
        self.mutation = mutation
        self.recombination = recombination
        self.random_step_prob = random_step_prob

    def sample(
        self,
        batch_size: int,
        model_states=None,
        env_states=None,
        walkers_states=None,
        **kwargs,
    ):
        """
        Calculate the corresponding data to interact with the Environment and \
        store it in model states.

        Args:
            batch_size: Number of new points to the sampled.
            model_states: SwarmState corresponding to the environment data.
            env_states: SwarmState corresponding to the model data.
            walkers_states: SwarmState corresponding to the walkers data.
            kwargs: Passed to the :class:`Critic` if any.

        Returns:
            Tuple containing a tensor with the sampled actions and the new model states variable.

        """
        # There is a chance of performing a gaussian perturbation
        if random_state.random() < self.random_step_prob:
            return super(ESModel, self).sample(
                batch_size=batch_size,
                env_states=env_states,
                model_states=model_states,
                **kwargs,
            )
        observs = (
            env_states.observs
            if env_states is not None
            else numpy.zeros(((batch_size,) + self.shape))
        )
        has_best = walkers_states is not None and walkers_states.best_state is not None
        best = walkers_states.best_state if has_best else observs
        # Choose 2 random indices
        indexes = numpy.arange(observs.shape[0])
        a_rand = self.random_state.permutation(indexes)
        b_rand = self.random_state.permutation(indexes)
        proposal = best + self.recombination * (observs[a_rand] - observs[b_rand])
        # Randomly mutate each coordinate of the original vector
        assert observs.shape == proposal.shape
        rands = random_state.random(observs.shape)
        perturbations = numpy.where(rands < self.mutation, observs, proposal)
        # The Environment will sum the observations to perform the step
        new_states = perturbations - observs
        actions = self.bounds.clip(new_states)
        return self.update_states_with_critic(
            actions=actions,
            batch_size=batch_size,
            model_states=model_states,
            env_states=env_states,
            walkers_states=walkers_states,
            **kwargs,
        )


class CMAES(Gaussian):
    """Implementation of CMAES algorithm from https://en.wikipedia.org/wiki/CMA-ES."""

    def __init__(self, sigma: float, virtual_reward_fitness: bool = False, *args, **kwargs):
        """
        Initialize a :class:`CMAES`.

        Args:
            sigma: Coordinate-wise standard deviation.
            virtual_reward_fitness: Use the virtual reward instead of the environment \
                                    reward to as a fitness measure.
            *args: Passed to :class:`NormalContinuous`.__init__.
            **kwargs: Passed to :class:`NormalContinuous`.__init__.

        """
        super(CMAES, self).__init__(*args, **kwargs)
        # Parameters for selection
        self.virtual_reward_fitness = virtual_reward_fitness
        self.pop_size = None
        self.sigma = sigma
        self._count_eval = 0  # Number of function evaluations performed
        self.x_mean = None  # Objective variables initial point
        self.old_x_mean = None  # x_mean corresponding to the last iteration
        # Parameters for adaptation that are kept constant
        self.mu_const = None  # Number of parents/points for recombination
        self.weights_const = None  # Array for weighted recombination
        self.mu_eff_const = None  # variance-effectiveness of sum w_i x_i
        self.cum_covm_const = None  # Time constant for accumulation for cov_matrix
        self.cum_sigma_const = None  # Time constant for accumulation for sigma control
        self.lr_covrank1_const = None  # Learning rate for rank-one update of cov_matrix
        self.lr_mu_const = None  # Learning rate for rank-mu_const update
        self.damp_sigma_const = None  # Damping for sigma. Usually close to one.
        self.chi_norm_const = None  # expectation of  ||N(0,I)|| == norm(randn(N,1))
        # Dynamic (internal) strategy parameters and constants
        self.paths_covm = None  # Evolution paths for cov_matrix
        self.paths_sigma = None  # Evolution paths for sigma
        self.coords_matrix = None  # Defines the coordinate system
        self.scaling_diag = None  # Diagonal matrix that defines the scaling
        self.cov_matrix = None  # Covariance matrix
        self.invsqrtC = None  # cov_matrix^-1/2
        self.n_eigen_eval = None  # Tracks update of coords_matrix and scaling_diag
        self.artmp = None  # Temporal array that contains terms for updating the covariance matrix
        self.hsig = None

    def sample(
        self,
        batch_size: int,
        model_states=None,
        env_states=None,
        walkers_states=None,
        **kwargs,
    ):
        """
        Calculate the corresponding data to interact with the Environment and \
        store it in model states.

        Args:
            batch_size: Number of new points to the sampled.
            model_states: SwarmState corresponding to the environment data.
            env_states: SwarmState corresponding to the model data.
            walkers_states: SwarmState corresponding to the walkers data.
            kwargs: Passed to the :class:`Critic` if any.

        Returns:
            Tuple containing a tensor with the sampled actions and the new model states variable.

        """
        if model_states is None or walkers_states is None:
            return super(CMAES, self).sample(
                batch_size=batch_size,
                model_states=model_states,
                env_states=env_states,
                walkers_states=walkers_states,
                **kwargs,
            )
        actions = (
            env_states.get("observs")
            if self._count_eval > self.pop_size * 2
            else model_states.get("actions")
        )
        fitness = (
            walkers_states.get("virtual_rewards")
            if self.virtual_reward_fitness
            else walkers_states.get("cum_rewards")
        )
        sorted_fitness = numpy.argsort(fitness)[: self.mu_const]
        selected_actions = actions[sorted_fitness].T
        self._update_evolution_paths(selected_actions)
        self._adapt_covariance_matrix(selected_actions)
        self._adapt_sigma()
        self._cov_matrix_diagonalization()

        actions = self._sample_actions()
        return self.update_states_with_critic(
            actions=actions,
            batch_size=batch_size,
            model_states=model_states,
            **kwargs,
        )

    def _update_evolution_paths(self, actions):
        self.old_x_mean = self.x_mean
        self.x_mean = numpy.matmul(actions, self.weights_const)
        # Update evolution paths
        self.paths_sigma = (1 - self.cum_sigma_const) * self.paths_sigma + numpy.sqrt(
            self.cum_sigma_const * (2 - self.cum_sigma_const) * self.mu_eff_const,
        ) * numpy.matmul(self.invsqrtC, (self.x_mean - self.old_x_mean)) / self.sigma
        sqrt_term = numpy.sqrt(
            1 - (1 - self.cum_sigma_const) ** (2 * self._count_eval / self.pop_size),
        )
        self.hsig = float(
            numpy.linalg.norm(self.paths_sigma) / sqrt_term / self.chi_norm_const
            < 1.4 + 2 / (self.n_dims + 1),
        )
        self.paths_covm = (1 - self.cum_covm_const) * self.paths_covm + self.hsig * numpy.sqrt(
            self.cum_covm_const * (2 - self.cum_covm_const) * self.mu_eff_const,
        ) * (self.x_mean - self.old_x_mean) / self.sigma

    def _adapt_covariance_matrix(self, actions):
        self.artmp = (1 / self.sigma) * (actions - numpy.tile(self.old_x_mean, (1, self.mu_const)))

        self.cov_matrix = (
            (1 - self.lr_covrank1_const - self.lr_mu_const) * self.cov_matrix
            + self.lr_covrank1_const * numpy.matmul(self.paths_covm, self.paths_covm.T)
            + (1 - self.hsig) * self.cum_covm_const * (2 - self.cum_covm_const) * self.cov_matrix
            + self.lr_mu_const
            * numpy.matmul(
                numpy.matmul(self.artmp, numpy.diag(self.weights_const.flatten())),
                self.artmp.T,
            )
        )

    def _adapt_sigma(self):
        self.sigma = self.sigma * numpy.exp(
            (self.cum_sigma_const / self.damp_sigma_const)
            * (numpy.linalg.norm(self.paths_sigma) / self.chi_norm_const - 1),
        )

    def _cov_matrix_diagonalization(self):
        # Decomposition of cov_matrix into coords_matrix*diag(scaling_diag.^2)*coords_matrix'
        # (diagonalization)
        if (
            self._count_eval - self.n_eigen_eval
            > self.pop_size / (self.lr_covrank1_const + self.lr_mu_const) / self.n_dims / 10
        ):
            self.n_eigen_eval = self._count_eval
            self.cov_matrix = numpy.triu(self.cov_matrix) + numpy.triu(self.cov_matrix, 1).T
            eigvals, eigvects = numpy.linalg.eig(self.cov_matrix)
            self.scaling_diag = numpy.diag(eigvals)  # [::-1])
            self.coords_matrix = eigvects  # [:, ::-1]
            self.scaling_diag = numpy.sqrt(numpy.diag(self.scaling_diag)).reshape(-1, 1)
            self.invsqrtC = numpy.matmul(
                numpy.matmul(self.coords_matrix, numpy.diag(self.scaling_diag.flatten() ** -1)),
                self.coords_matrix.T,
            )

    def reset(
        self,
        batch_size: int = 1,
        model_states=None,
        env_states=None,
        *args,
        **kwargs,
    ):
        """
        Return a new blank State for a `DiscreteUniform` instance, and a valid \
        prediction based on that new state.

        Args:
            batch_size: Number of walkers that the new model `State`.
            model_states: :class:`StatesModel` corresponding to the model data.
            env_states: :class:`StatesEnv` containing the environment data.
            *args: Passed to `predict`.
            **kwargs: Passed to `predict`.

        Returns:
            New model states containing sampled data.

        """
        self.pop_size = batch_size
        self._count_eval = 0
        self._init_algorithm_params(batch_size)
        # Take the first sample from a random uniform distribution
        if batch_size is None and env_states is None:
            raise ValueError("env_states and batch_size cannot be both None.")
        batch_size = batch_size or env_states.n
        model_states = model_states or self.create_new_states(batch_size=batch_size)
        init_actions = self.random_state.randn(self.mu_const)
        self.x_mean = numpy.matmul(init_actions.T, self.weights_const)
        actions = self._sample_actions()
        model_states.update(actions=actions)
        return model_states

    def _sample_actions(self) -> numpy.ndarray:

        actions = numpy.empty((self.pop_size, self.n_dims))
        for i in range(self.pop_size):
            normal_noise = self.random_state.randn(self.n_dims).reshape(-1, 1)
            action = self.x_mean + self.sigma * numpy.matmul(
                self.coords_matrix,
                self.scaling_diag * normal_noise,
            )
            actions[i, :] = action.flatten()
            self._count_eval += 1
        return actions

    def _init_algorithm_params(self, batch_size):
        self.pop_size = batch_size
        self.mu_const = self.pop_size / 2  # Number of parents/points for recombination
        self.weights_const = numpy.log(self.mu_const + 0.5) - numpy.log(
            numpy.arange(1, self.mu_const + 1),
        ).reshape((-1, 1))
        self.mu_const = int(numpy.floor(self.mu_const))
        self.weights_const = self.weights_const / numpy.sum(self.weights_const)
        self.mu_eff_const = numpy.sum(self.weights_const) ** 2 / numpy.sum(self.weights_const**2)
        # Parameters for adaptation
        self.cum_covm_const = (4 + self.mu_eff_const / self.n_dims) / (
            self.n_dims + 4 + 2 * self.mu_eff_const / self.n_dims
        )
        self.cum_sigma_const = (self.mu_eff_const + 2) / (self.n_dims + self.mu_eff_const + 5)
        self.lr_covrank1_const = 2 / ((self.n_dims + 1.3) ** 2 + self.mu_eff_const)
        self.lr_mu_const = numpy.minimum(
            1 - self.lr_covrank1_const,
            2
            * (self.mu_eff_const - 2 + 1 / self.mu_eff_const)
            / ((self.n_dims + 2) ** 2 + self.mu_eff_const),
        )
        self.damp_sigma_const = (
            1
            + 2 * numpy.maximum(0, numpy.sqrt((self.mu_eff_const - 1) / (self.n_dims + 1)) - 1)
            + self.cum_sigma_const
        )
        # Dynamic (internal) strategy parameters and constants
        self.paths_covm = numpy.zeros((self.n_dims, 1))
        self.paths_sigma = numpy.zeros((self.n_dims, 1))
        self.coords_matrix = numpy.eye(self.n_dims)  # Defines the coordinate system
        self.scaling_diag = numpy.ones(
            (self.n_dims, 1),
        )  # Diagonal matrix that defines the scaling
        self.cov_matrix = (
            self.coords_matrix * numpy.diag(self.scaling_diag**2) * self.coords_matrix.T
        )  # Covariance matrix
        self.invsqrtC = (
            self.coords_matrix * numpy.diag(1 / self.scaling_diag.flatten()) * self.coords_matrix.T
        )
        self.n_eigen_eval = 0  # Tracks update of coords_matrix and scaling_diag
        self.chi_norm_const = self.n_dims**0.5 * (
            1 - 1 / (4 * self.n_dims) + 1 / (21 * self.n_dims**2)
        )



from typing import Union

import judo
from judo.data_types import dtype
from judo.functions.random import random_state
import numpy

from fragile.core.api_classes import Callback
from fragile.core.typing import StateData, Tensor


class TimeStepAPI(Callback):
    name = "time_step"
    default_param_dict = {"dt": {"dtype": dtype.int32}}
    default_outputs = tuple(["dt"])

    def select(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        """Calculate SwarmState containing the data needed to interact with the environment."""
        dt_input = self._prepare_tensors(**kwargs)
        dt_data = self.calculate(**dt_input)
        if not isinstance(dt_data, dict):
            dt_data = {"dt": dt_data}
        if inplace:
            self.update(**dt_data)
        else:
            return dt_data

    def calculate(self, **kwargs) -> Tensor:
        raise NotImplementedError

    def reset(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        return self.select(inplace=inplace)

    def before_policy(self):
        self.select()


class ConstantDt(TimeStepAPI):
    def __init__(self, value: int, **kwargs):
        super(ConstantDt, self).__init__(**kwargs)
        self.value = value

    def calculate(self) -> Tensor:
        return self.value * judo.ones(self.swarm.n_walkers, dtype=dtype.int32)


class UniformDt(TimeStepAPI):
    def __init__(self, high: int, low: int = 1, **kwargs):
        super(UniformDt, self).__init__(**kwargs)
        self.low = low
        self.high = high

    def calculate(self) -> Tensor:
        return random_state.randint(
            low=self.low,
            high=self.high,
            size=self.swarm.n_walkers,
            dtype=dtype.int32,
        )

    def reset(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        dt_data = {"dt": self.low * judo.ones(self.swarm.n_walkers, dtype=dtype.int32)}
        if inplace:
            self.update(**dt_data)
        else:
            return dt_data


class GaussianDt(UniformDt):
    def __init__(self, loc: float, scale: float, low: int = 1, high=numpy.inf, **kwargs):
        self.loc = loc
        self.scale = scale
        super(GaussianDt, self).__init__(high=high, low=low, **kwargs)

    def calculate(self) -> Tensor:
        dts = random_state.normal(loc=self.loc, scale=self.scale, size=self.n_walkers)
        return judo.astype(judo.clip(dts, self.low, self.high), dtype=dtype.int32)



import judo
from judo.data_types import dtype

from fragile.core.api_classes import Callback
from fragile.core.typing import StateDict


class TrackSteps(Callback):
    name = "track_steps"
    default_inputs = {"n_step": {"clone": True}}
    default_outputs = ("n_step",)

    @property
    def param_dict(self) -> StateDict:
        return {"n_step": {"dtype": dtype.int32}}

    def update_steps(self) -> None:
        actives = self.swarm.state.actives
        infos = self.swarm.state.get("infos", inactives=True)
        if infos is None:
            return
        new_steps = judo.tensor([info.get("n_step") for info in infos], dtype=dtype.int32)
        steps = self.swarm.state.get("n_step", inactives=True)
        steps[actives] = steps[actives] + new_steps[actives]
        self.swarm.state.update(n_step=steps, inactives=True)

    def after_env(self):
        self.update_steps()



import copy

import judo
import numpy

from fragile.core.api_classes import Callback


class RootWalker(Callback):
    name = "root"

    def __init__(self, **kwargs):
        self._data = {}
        self.minimize = False
        super(RootWalker, self).__init__(**kwargs)

    def __getattr__(self, item):
        plural = item + "s"
        if plural in self._data:
            d = self._data[plural]
            try:
                return (
                    d[0]
                    if isinstance(d, (list, numpy.ndarray))
                    else (d.item() if len(d.shape) == 0 else d)
                )
            except IndexError:
                return d.item()

        elif item in self._data:
            d = self._data[item]
            try:
                return (
                    d[0]
                    if isinstance(d, (list, numpy.ndarray))
                    else (d.item() if len(d.shape) == 0 else d)
                )
            except IndexError:
                return d.item()

            return self._data[item][0]
        return self.__getattribute__(item)

    def __repr__(self) -> str:
        # score = self.data.get('scores', [numpy.nan])[0]
        return f"{self.__class__.__name__}: score: {self.scores}"

    def to_html(self):
        return (
            f"<strong>{self.__class__.__name__}</strong>: "
            f"Score: {self.scores}\n"
            # f"Score: {self.data.get('scores', [numpy.nan])[0]}\n"
        )

    @property
    def data(self):
        return self._data

    def setup(self, swarm):
        super(RootWalker, self).setup(swarm)
        self.minimize = swarm.minimize

    def reset(self, root_walker=None, state=None, **kwargs):
        if root_walker is None:
            value = [numpy.inf if self.minimize else -numpy.inf]
            self._data = {"scores": value, "rewards": value}
            self.update_root()
        else:
            self._data = {k: copy.deepcopy(v) for k, v in root_walker.items()}

    def before_walkers(self):
        self.update_root()

    def update_root(self):
        raise NotImplementedError()


class BestWalker(RootWalker):
    default_inputs = {"scores": {}, "oobs": {"optional": True}}

    def __init__(self, always_update: bool = False, fix_root=True, **kwargs):
        super(BestWalker, self).__init__(**kwargs)
        self.minimize = None
        self.always_update = always_update
        self._fix_root = fix_root

    def get_best_index(self):
        scores, oobs, terminals = self.get("scores"), self.get("oobs"), self.get("terminals")
        index = judo.arange(len(scores))
        bool_ix = ~oobs if terminals is None else judo.logical_or(~oobs, terminals)
        alive_scores = scores[bool_ix]
        if len(alive_scores) == 0:
            return 0
        ix = alive_scores.argmin() if self.minimize else alive_scores.argmax()
        return index[bool_ix][ix]

    def get_best_walker(self):
        return self.swarm.state.export_walker(self.get_best_index())

    def update_root(self):
        best = self.get_best_walker()
        best_score = best["scores"][0] if judo.Backend.is_numpy() else best["scores"].item()

        # if not judo.Backend.is_numpy():# and judo.dtype.is_tensor(self.score):
        #    scores = self._data["scores"]
        #    score = scores[0] if isinstance(scores, list) else scores.item()
        # else:
        score = self.score
        score_improves = (best_score < score) if self.minimize else (best_score > score)
        if self.always_update or score_improves:  # or numpy.isinf(score):
            # new_best = {k: copy.deepcopy(v) for k, v in best.items()}
            self._data = copy.deepcopy(best)

    def fix_root(self):
        if self._fix_root:
            self.swarm.state.import_walker(copy.deepcopy(self.data))
            terminals = self.swarm.get("terminals")
            if not self.swarm.state.actives[0] and terminals is not None and not terminals[0]:
                self.swarm.state.actives[0] = True
                self.swarm.state._n_actives += 1

    def after_walkers(self):
        self.fix_root()


class TrackWalker(RootWalker):
    default_inputs = {"scores": {}, "oobs": {"optional": True}}

    def __init__(self, walker_index=0, **kwargs):
        super(TrackWalker, self).__init__(**kwargs)
        self.walker_index = walker_index

    def update_root(self):
        walker = self.swarm.state.export_walker(self.walker_index)
        self._data = copy.deepcopy({k: copy.deepcopy(v) for k, v in walker.items()})



# import holoviews
from holoviews.streams import Pipe  # Buffer
import pandas
import panel

from fragile.dataviz.stream_plots import (  # Histogram,; Landscape2D,; QuadMeshContours,
    PlotCallback,
    Scatter,
)


class Plot2DSwarm(PlotCallback):
    name = "plot_2d"

    def __init__(self, **kwargs):
        super(Plot2DSwarm, self).__init__(**kwargs)
        stream = Pipe(data=pandas.DataFrame({"x": [0], "y": [0], "c": [0]}))

        self.walkers_scatter_sp = Scatter(
            title="Score Landscape",
            stream=stream,
            color="c",
            cmap="viridis",
            colorbar=True,
            # plot=lambda data: holoviews.Scatter(data).opts(color="c")
        )
        # self.landscape_2d_sp = QuadMeshContours()

    def setup(self, swarm):
        super(Plot2DSwarm, self).setup(swarm)
        # bounds = swarm.env.bounds
        # xlim, ylim = (bounds.low[0], bounds.high[0]), (bounds.low[1], bounds.high[1])
        # self.landscape_2d_sp.plot = self.landscape_2d_sp.plot.redim(
        #    x=holoviews.Dimension("x", range=xlim),
        #    y=holoviews.Dimension("y", range=ylim),
        # )

    def send(self):
        observs = self.get("observs")
        scores = self.get("scores")
        df = pandas.DataFrame({"x": observs[:, 0], "y": observs[:, 1], "c": scores})
        # data = (observs[:, 0], observs[:, 1], scores)
        self.walkers_scatter_sp.send(df)
        # self.landscape_2d_sp.send(data)

    def panel(self):
        return panel.Column(self.walkers_scatter_sp.plot)  # + self.landscape_2d_sp.plot)



import logging
from typing import Optional

import einops
from judo.functions.notebook import running_in_ipython
import numpy
from tqdm.autonotebook import tqdm

from fragile.core.api_classes import Callback


def statistics_from_array(x: numpy.ndarray):
    """Return the (mean, std, max, min) of an array."""
    try:
        return (
            einops.asnumpy(x).mean(),
            einops.asnumpy(x).std(),
            einops.asnumpy(x).max(),
            einops.asnumpy(x).min(),
        )
    except (AttributeError, TypeError, ValueError):
        return numpy.nan, numpy.nan, numpy.nan, numpy.nan


class Report(Callback):
    name = "report"
    _log = logging.getLogger("Swarm")

    def __init__(
        self,
        report_interval: Optional[int] = None,
        progress_bar: bool = True,
        notebook_widget: bool = True,
        **kwargs,
    ):
        self._use_progress_bar = progress_bar
        self._use_notebook_widget = notebook_widget
        self._ipython_mode = None
        self._notebook_container = None
        self.report_interval = report_interval
        self.tqdm = None
        self.html_output = None
        super(Report, self).__init__(**kwargs)

    def before_reset(self):
        self._ipython_mode = running_in_ipython()
        if self._use_progress_bar:
            self.tqdm = tqdm(total=self.swarm.max_epochs)
        if self._use_notebook_widget:
            self.setup_notebook_container()

    def after_evolve(self):
        if self._use_progress_bar:
            self.tqdm.update(1)
        if self.report_interval is not None and self.swarm.epoch % self.report_interval == 0:
            self.report_progress()

    def run_end(self):
        if self._use_progress_bar:
            self.tqdm.close()
        if self.report_interval is not None:
            self.report_progress()

    def report_progress(self):
        """Report information of the current run."""
        if self._ipython_mode and self._use_notebook_widget:
            html = self.swarm.to_html()
            self._notebook_container.value = "%s" % html
        elif not self._ipython_mode:
            self._log.info(repr(self))

    def setup_notebook_container(self):
        """Display the display widgets if the Swarm is running in an IPython kernel."""
        if self._ipython_mode and self._use_notebook_widget:
            from IPython.core.display import display, HTML as cell_html
            from ipywidgets import HTML

            # Set font weight of tqdm progressbar
            display(cell_html("<style> .widget-label {font-weight: bold !important;} </style>"))
            self._notebook_container = HTML()
            display(self._notebook_container)



from typing import Optional, Union

import judo
from judo import Backend, Bounds, dtype, random_state, tensor
import numpy.linalg

from fragile.core.api_classes import PolicyAPI, SwarmAPI
from fragile.core.typing import StateData, Tensor


class RandomPlangym(PolicyAPI):
    """Policy that selects random actions from the environment action space."""

    def setup(self, swarm: SwarmAPI):
        """
        Setup the policy.

        Args:
            swarm (SwarmAPI): Swarm that will use this policy.

        Raises:
            TypeError: If the environment does not have a sample_action method or an action_space.
        """
        super(RandomPlangym, self).setup(swarm)
        if hasattr(self.swarm.env, "sample_action"):
            sample = self.swarm.env.sample_action
        elif hasattr(self.swarm.env, "action_space"):
            sample = self.swarm.env.action_space.select_dt
        elif hasattr(self.swarm.env, "plangym_env"):
            penv = self.swarm.env.plangym_env
            if hasattr(penv, "sample_action"):
                sample = penv.sample_action
            elif hasattr(penv, "action_space"):
                sample = penv.action_space.select_dt
        else:
            raise TypeError("Environment does not have a sample_action method or an action_space")
        self._sample_func = sample

    def select_actions(self, **kwargs) -> list:
        """Sample a random action from the environment action space."""
        a = [self._sample_func() for _ in range(len(self.swarm))]
        return a


class Discrete(PolicyAPI):
    """A policy that selects discrete actions according to some probability distribution."""

    def __init__(self, actions: Optional[Tensor] = None, probs: Optional[Tensor] = None, **kwargs):
        """
        Initialize a :class:`Discrete` policy.

        Args:
            actions (Optional[Tensor], optional): The possible actions that can be
                taken by this policy. Defaults to None.. Defaults to None.
            probs (Optional[Tensor], optional): The probabilities of selecting each action.
                Not needed for uniform distributions. Defaults to None.. Defaults to None.
            **kwargs: Any other values that should be set for the policy.
        """
        super(Discrete, self).__init__(**kwargs)
        # TODO: parse all possible ways to infer actions, probs, and n_actions
        self.probs = probs
        self._n_actions = None
        self._actions = actions
        self._setup_params(actions)

    @property
    def n_actions(self) -> int:
        """The number of possible actions."""
        return self._n_actions

    @property
    def actions(self) -> Tensor:
        """The possible actions that can be taken by this policy."""
        return self._actions

    def select_actions(self, **kwargs) -> Tensor:
        """
        Select a random action from the possible actions.

        Returns:
           An array of shape (swarm.n_walkers,) containing the selected actions.
        """
        return random_state.choice(self.actions, p=self.probs, size=self.swarm.n_actives)

    def setup(self, swarm: SwarmAPI) -> None:
        """
        Sets up the policy, inferring any missing parameters.

        Args:
            swarm (SwarmAPI): The swarm that is using this policy.

        Raises:
            TypeError: If n_actions cannot be inferred.
        """
        self._swarm = swarm
        if self.n_actions is None:
            if hasattr(self.swarm.env, "n_actions"):
                self._n_actions = self.swarm.env.n_actions
            elif hasattr(self.swarm.env, "action_space"):
                self._n_actions = self.swarm.env.action_space.n
            else:
                raise TypeError("n_actions cannot be inferred.")
        self._setup_params(self.actions)

    def _setup_params(self, actions: Optional[Tensor] = None):
        """Setup the parameters of the policy."""
        if actions is None and self.n_actions is None and self.actions is not None:
            self._n_actions = len(self.actions)  # Try to set up n_actions using actions
        elif actions is None and self.n_actions is not None:  # setup actions with n_actions
            self._actions = judo.arange(self.n_actions) if self.actions is None else self.actions
        elif isinstance(actions, (list, tuple)) or judo.is_tensor(actions):
            self._actions = tensor(actions)
            self._n_actions = len(self.actions)
        elif actions is not None:  # Actions is an integer-like value
            self._n_actions = self.n_actions if self.n_actions is not None else int(actions)
            self._actions = judo.arange(self._n_actions)
        elif self.probs is not None:  # Try to infer values from probs
            self._n_actions = len(self.probs)
            self._actions = judo.arange(self._n_actions)


class BinarySwap(PolicyAPI):
    def __init__(self, n_swaps: int, n_actions: int = None, **kwargs):
        self._n_actions = n_actions
        self._n_swaps = n_swaps
        super(BinarySwap, self).__init__(**kwargs)

    @property
    def n_actions(self):
        return self._n_actions

    @property
    def n_swaps(self):
        return self._n_swaps

    def select_actions(self, **kwargs):
        from numba import jit
        import numpy

        @jit(nopython=True)
        def flip_values(actions: numpy.ndarray, flips: numpy.ndarray):
            for i in range(flips.shape[0]):
                for j in range(flips.shape[1]):
                    actions[i, flips[i, j]] = numpy.logical_not(actions[i, flips[i, j]])
            return actions

        observs = judo.to_numpy(self.get("observs"))
        with Backend.use_backend("numpy"):
            actions = judo.astype(observs, dtype.bool)
            flips = random_state.randint(0, self.n_actions, size=(observs.shape[0], self.n_swaps))
            actions = judo.astype(flip_values(actions, flips), dtype.int64)
        actions = tensor(actions)
        return actions

    def setup(self, swarm: SwarmAPI):
        self._swarm = swarm
        if self.n_actions is None:
            if hasattr(self.swarm.env, "n_actions"):
                self._n_actions = self.swarm.env.n_actions
            elif hasattr(self.swarm.env, "action_space"):
                self._n_actions = self.swarm.env.action_space.n
            else:
                raise TypeError("n_actions cannot be inferred.")


class ContinuousPolicy(PolicyAPI):
    """
    ContinuousPolicy implements a continuous action space policy for interacting \
    with the environment.

    Args:
        bounds (Bounds, optional): Action space bounds. If not provided, the bounds are obtained
            from the environment. Defaults to `None`.
        second_order (bool, optional): If `True`, the policy is considered second-order, and the \
            action sampled will be added to the last value. Defaults to `False`.
        step (float, optional): The step size for updating the actions. Defaults to 1.0.
        **kwargs: Additional keyword arguments for the base PolicyAPI class.

    Attributes:
        second_order (bool): If `True`, the policy is considered second-order.
        step (float): The step size for updating the actions.
        bounds (Bounds): Action space bounds.
        _env_bounds (Bounds): Environment action space bounds.
    """

    def __init__(self, bounds=None, second_order: bool = False, step: float = 1.0, **kwargs):
        """Initialize a :class:`ContinuousPolicy`."""
        self.second_order = second_order
        self.step = step
        self.bounds = bounds
        self._env_bounds = None
        if second_order:
            kwargs["inputs"] = {"actions": {"clone": True}, **kwargs.get("inputs", {})}
        super(ContinuousPolicy, self).__init__(**kwargs)

    @property
    def env_bounds(self) -> Bounds:
        """Returns the environment action space bounds."""
        return self._env_bounds

    def select_actions(self, **kwargs):
        """
        Implement the functionality for selecting actions in the derived class. This method is
        called during the act operation.

        Args:
            **kwargs: Additional keyword arguments required for selecting actions.
        """
        raise NotImplementedError()

    def act(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        """
        Calculate the data needed to interact with the :class:`Environment`.

        Args:
            inplace (bool, optional): If `True`, updates the swarm state with the selected actions.
                                      If `False`, returns the selected actions. Defaults to `True`.
            **kwargs: Additional keyword arguments required for acting.

        Returns:
            Union[None, StateData]: A dictionary containing the selected actions if inplace is
                                    `False. Otherwise, returns `None`.
        """
        action_input = self._prepare_tensors(**kwargs)
        actions_data = self.select_actions(**action_input)
        if not isinstance(actions_data, dict):
            actions_data = {"actions": actions_data}
        actions = actions_data["actions"]
        if self.second_order:
            prev_actions = action_input["actions"]
            actions = prev_actions + actions * self.step
        actions_data["actions"] = self.env_bounds.clip(actions)
        if inplace:
            self.update(**actions_data)
        else:
            return actions_data

    def setup(self, swarm: SwarmAPI) -> None:
        """
        Set up the policy with the provided swarm object.

        Args:
            swarm (SwarmAPI): The swarm object to set up the policy with.

        Returns:
            None
        """
        super(ContinuousPolicy, self).setup(swarm)
        if self.bounds is None:
            if hasattr(self.swarm.env, "bounds"):
                self.bounds = self.swarm.env.bounds
            elif hasattr(self.swarm.env, "action_space"):
                self.bounds = Bounds.from_space(self.swarm.env.action_space)
            else:
                raise ValueError("Bounds is not defined and not present in the Environment.")
        if hasattr(self.swarm.env, "bounds"):
            self._env_bounds = self.swarm.env.bounds
        elif hasattr(self.swarm.env, "action_space"):
            self._env_bounds = Bounds.from_space(self.swarm.env.action_space)
        else:
            self._env_bounds = self.bounds


class ZeroContinuous(ContinuousPolicy):
    """
    Uniform policy samples actions equal to zero.

    Inherits from :class:`ContinuousPolicy`.
    """

    def select_actions(self, **kwargs) -> Tensor:
        """
        Select a vector of zeros.

        Args:
            **kwargs: Additional keyword arguments required for selecting actions.

        Returns:
            Tensor: Selected actions as a tensor.
        """
        shape = tuple([self.swarm.n_actives]) + self.bounds.shape
        return judo.zeros(shape, dtype=self.bounds.dtype)


class Uniform(ContinuousPolicy):
    """
    Uniform policy samples actions uniformly from the bounds of the action space.

    Inherits from :class:`ContinuousPolicy`.
    """

    def select_actions(self, **kwargs) -> Tensor:
        """
        Select actions by sampling uniformly from the action space bounds.

        Args:
            **kwargs: Additional keyword arguments required for selecting actions.

        Returns:
            Tensor: Selected actions as a tensor.
        """
        shape = tuple([self.swarm.n_actives]) + self.bounds.shape
        new_points = random_state.uniform(
            low=self.bounds.low,
            high=self.bounds.high,
            size=shape,
        )
        return new_points


class Gaussian(ContinuousPolicy):
    """
    The Gaussian policy samples actions from a Gaussian distribution.

    Inherits from :class:`ContinuousPolicy`.
    """

    def __init__(self, loc: float = 0.0, scale: float = 1.0, **kwargs):
        """
        Initialize a :class:`Gaussian` policy.

        Args:
            loc (float, optional): Mean of the policy. Defaults to 0.0.
            scale (float, optional): Standard deviation of the policy. Defaults to 1.0.
            **kwargs: Additional arguments.
        """
        super(Gaussian, self).__init__(**kwargs)
        self.loc = loc
        self.scale = scale

    def select_actions(self, **kwargs) -> Tensor:
        """
        Select actions by sampling from a Gaussian distribution.

        Args:
            **kwargs: Additional keyword arguments required for selecting actions.

        Returns:
            Tensor: Selected actions as a tensor.
        """
        shape = tuple([self.swarm.n_actives]) + self.bounds.shape
        new_points = random_state.normal(
            loc=self.loc,
            scale=self.scale,
            size=shape,
        )
        return new_points


class GaussianModulus(ContinuousPolicy):
    def __init__(self, loc: float = 0.0, scale: float = 1.0, **kwargs):
        super(GaussianModulus, self).__init__(**kwargs)
        self.loc = loc
        self.scale = scale

    def select_actions(self, **kwargs) -> Tensor:
        shape = tuple([self.swarm.n_actives]) + self.bounds.shape
        new_points = random_state.uniform(
            low=-1.0,
            high=1.0,
            size=shape,
        )
        new_points = new_points / numpy.linalg.norm(new_points, axis=1).reshape(-1, 1)

        modulus = random_state.normal(
            loc=self.loc,
            scale=self.scale,
            size=(self.swarm.n_actives, 1),
        )

        return new_points * modulus



import copy
from typing import Callable, Optional

import einops
from judo.data_structures.bounds import Bounds
import numpy
import pandas

from fragile.core.api_classes import SwarmAPI
from fragile.core.env import Function
from fragile.core.state import SwarmState


class Swarm(SwarmAPI):
    def __repr__(self):
        text = f"{self.__class__.__name__}: walkers {self.n_walkers} iteration {self.epoch}\n"
        with numpy.printoptions(linewidth=100, threshold=200, edgeitems=9):
            if hasattr(self, "root"):
                text += self.root.__repr__()
            text += f"\nState statistics:\n{self._state_stats_df().round(3)}"
        return text

    def to_html(self):
        line_break = '<br style="line-height:1px; content: "  ";>'
        text = (
            f"<strong>{self.__class__.__name__}</strong>: "
            f"walkers {self.n_walkers} iteration {self.epoch}\n"
        )
        with numpy.printoptions(linewidth=100, threshold=200, edgeitems=9):
            if hasattr(self, "root"):
                text += self.root.to_html()
            if hasattr(self, "tree"):
                text += self.tree.to_html()
            text += "\n<strong>State statistics</strong>:\n"
            text = text.replace("\n\n", "\n").replace("\n", line_break)
            text += f"\n{self._state_stats_df().round(3).to_html()}"
        return text

    def _state_stats_df(self):
        skip_print = {"observs", "states", "id_walkers", "parent_ids", "infos", "compas_clone"}
        data = {
            k: einops.asnumpy(v)
            for k, v in self.state.items()
            if k in self.state.vector_names and k not in skip_print
        }
        return pandas.DataFrame(data).astype(float).describe().T.drop(columns=["count"])

    def _setup_inputs(self):
        # TODO: Make it so it does not override params
        inputs = {
            **copy.deepcopy(self.env.inputs),
            **copy.deepcopy(self.policy.inputs),
            **copy.deepcopy(self.walkers.inputs),
        }
        for callback in self.callbacks.values():
            inputs.update(dict(callback.inputs))
        self._inputs = inputs

    def _setup_clone_names(self):
        clone_env = [k for k, v in self.env.inputs.items() if v.get("clone")]
        clone_policy = [k for k, v in self.policy.inputs.items() if v.get("clone")]
        clone_walkers = [k for k, v in self.walkers.inputs.items() if v.get("clone")]
        clone_names = clone_env + clone_policy + clone_walkers
        for callback in self.callbacks.values():
            clone_callback = [k for k, v in callback.inputs.items() if v.get("clone")]
            clone_names = clone_names + clone_callback
        self._clone_names = set(list(self.clone_names) + clone_names)

    def _setup_components(self):
        self.env.setup(self)
        self.setup_state(n_walkers=self.n_walkers, param_dict=self.env.param_dict)
        self.policy.setup(self)
        acc_params = {**self.env.param_dict, **self.policy.param_dict, **self.state.param_dict}
        self.setup_state(n_walkers=self.n_walkers, param_dict=acc_params)
        self.walkers.setup(self)
        acc_params = {
            **self.env.param_dict,
            **self.policy.param_dict,
            **self.walkers.param_dict,
            **self.state.param_dict,
        }
        self.setup_state(n_walkers=self.n_walkers, param_dict=acc_params)
        for callback in self.callbacks.values():
            callback.setup(self)
            acc_params.update(callback.param_dict)
            self.setup_state(n_walkers=self.n_walkers, param_dict=acc_params)


class FunctionMapper(Swarm):
    """It is a swarm adapted to minimize mathematical functions."""

    def __init__(
        self,
        minimize: bool = True,
        start_same_pos: bool = False,
        **kwargs,
    ):
        """
        Initialize a :class:`FunctionMapper`.

        Args:
            accumulate_rewards: If `True` the rewards obtained after transitioning
                to a new state will accumulate. If `False` only the last reward
                will be taken into account.
            minimize: If `True` the algorithm will perform a minimization process.
                If `False` it will be a maximization process.
            start_same_pos: If `True` all the walkers will have the same starting position.
            **kwargs: Passed :class:`Swarm` __init__.
        """
        super(FunctionMapper, self).__init__(
            minimize=minimize,
            **kwargs,
        )
        self.start_same_pos = start_same_pos

    @classmethod
    def from_function(
        cls,
        function: Callable,
        bounds: Bounds,
        **kwargs,
    ) -> "FunctionMapper":
        """
        Initialize a :class:`FunctionMapper` using a python callable and a \
        :class:`Bounds` instance.

        Args:
            function: Callable representing an arbitrary function to be optimized.
            bounds: Represents the domain of the function to be optimized.
            **kwargs: Passed to :class:`FunctionMapper` __init__.

        Returns:
            Instance of :class:`FunctionMapper` that optimizes the target function.

        """
        env = Function(function=function, bounds=bounds)
        return FunctionMapper(env=env, **kwargs)

    def __repr__(self):
        return "{}\n{}".format(self.env.__repr__(), super(FunctionMapper, self).__repr__())

    def reset(
        self,
        root_walker: Optional["OneWalker"] = None,
        state: Optional[SwarmState] = None,
    ):
        """
        Reset the :class:`fragile.Walkers`, the :class:`Function` environment, the \
        :class:`Model` and clear the internal data to start a new search process.

        Args:
            root_walker: Walker representing the initial state of the search. \
                         The walkers will be reset to this walker, and it will \
                         be added to the root of the :class:`StateTree` if any.
            state: StateData dictionary that define the initial state of the Swarm.

        """
        super(FunctionMapper, self).reset(root_walker=root_walker, state=state)
        if self.start_same_pos:
            observs = self.get("observs")
            observs[:] = observs[0]
            self.update(observs=observs)
            states = self.get("states", None, raise_error=False)
            if states is not None:
                states[:] = states[0]
                self.update(states=states)



import copy
from typing import Any, Dict, Iterable, Optional, Set, Tuple, Union

import judo
from judo import dtype, random_state

from fragile.core.state import SwarmState
from fragile.core.typing import InputDict, StateData, StateDict, Tensor


class SwarmComponent:
    """
    A component of a swarm simulation.

    Every class that stores its data in :class:`SwarmState` inherits  from this class.

    Args:
        swarm: Reference to the :class:`Swarm` that incorporates this
            :class:`SwarmComponent`. Defaults to None.
        param_dict: Describes the multi-dimensional arrays
            that define the internal state of the component. Defaults to None.
        inputs: Describes the data that this components
            processes as inputs. This data is present in the internal state of the
            :class:`Swarm`, and generated by other components. Defaults to None.
        outputs: Name of the data attributes that this
            component makes available as part of its internal state. Defaults to None.

    Attributes:
        swarm (Swarm): The parent :class:`Swarm` instance.
        n_walkers (int): Number of walkers in the :class:`Swarm`.
        inputs (Dict[str, Dict[str, Any]]): A dictionary mapping input names
            to their corresponding type and default values.
        outputs (Tuple[str]): A tuple containing the names of the data attributes.
        param_dict (Dict[str, Dict[str, Any]]): A dictionary mapping parameter names.

    Methods:
        get(...)
        get_input_data(...)
        update(...)
        reset(...)
        _prepare_tensors(...)
    """

    default_inputs = {}
    default_outputs = tuple()
    default_param_dict = {}

    def __init__(
        self,
        swarm: Optional["SwarmAPI"] = None,
        param_dict: Optional[StateDict] = None,
        inputs: Optional[InputDict] = None,
        outputs: Optional[Tuple[str]] = None,
    ):
        """
        Initialize a :class:`SwarmComponent`.

        Args:
            swarm (Optional["SwarmAPI";], optional): Reference to the :class:`Swarm`
                that incorporates this :class:`SwarmComponent`. Defaults to None.
            param_dict (Optional[StateDict], optional): Describes the multi-dimensional arrays
                that define the internal state of the component. Defaults to None.
            inputs (Optional[InputDict], optional): Describes the data that this components
                processes as inputs. This data is present in the internal state of the
                :class:`Swarm`, and generated by other components. Defaults to None.
            outputs (Optional[Tuple[str]], optional): Name of the data attributes that this
                component makes available as part of its internal state. Defaults to None.
        """
        param_dict = param_dict or {}
        param_dict = {**self.default_param_dict, **param_dict}
        inputs = inputs or {}
        inputs = {**self.default_inputs, **inputs}
        outputs = tuple(outputs) if outputs is not None else tuple()
        outputs = tuple(set(self.default_outputs + outputs))
        self._swarm = None
        self._param_dict = param_dict
        self._inputs = inputs
        self._outputs = outputs
        if swarm is not None:  # This way you can run side effects on child classes
            self.setup(swarm)

    @property
    def swarm(self) -> "SwarmAPI":
        """Return a reference to the :class:`Swarm` that includes the current component."""
        return self._swarm

    @property
    def n_walkers(self) -> int:
        """Return the number of walkers of the :class:`Swarm`."""
        return self.swarm.n_walkers

    @property
    def inputs(self) -> InputDict:
        """Return a dictionary containing the data that this component needs to function."""
        return dict(self._inputs)

    @property
    def outputs(self) -> Tuple[str, ...]:
        """Return a tuple containing the names of the data attribute that the component outputs."""
        return tuple(self._outputs)

    @property
    def param_dict(self) -> StateDict:
        """Return the dictionary defining all the data attributes that the component requires."""
        return dict(self._param_dict)

    def setup(self, swarm):
        """Prepare the component during the setup phase of the :class:`Swarm`."""
        self._swarm = swarm

    def get(
        self,
        name: str,
        default: Any = None,
        raise_error: bool = False,
        inactives: bool = False,
    ) -> Any:
        """Access attributes of the :class:`Swarm` and its children."""
        return self.swarm.state.get(
            name=name,
            default=default,
            raise_error=raise_error,
            inactives=inactives,
        )

    def get_input_data(self) -> StateData:
        """
        Return a dictionary with all the data that this component requires as inputs.

        Returns:
            Dictionary containing all the data required by the :class:`SwarmComponent`,
            where each key is the name of a data attribute and its value
            a multi-dimensional array.
        """

        def get_one_input(name, values):
            raise_error = not values.get("optional", False)
            return self.get(name=name, default=values.get("default"), raise_error=raise_error)

        return {k: get_one_input(k, v) for k, v in self.inputs.items()}

    def update(
        self,
        other: Union["SwarmState", Dict[str, Tensor]] = None,
        inactives: bool = False,
        **kwargs,
    ) -> None:
        """
        Modify the data stored in the SwarmState instance.

        Existing attributes will be updated, and new attributes will be created if needed.

        Args:
            other: State class that will be copied upon update.
            inactives: Whether to update the walkers marked as inactive.
            **kwargs: It is possible to specify the update as name value attributes,
                where name is the name of the attribute to be updated, and value
                is the new value for the attribute.
        Returns:
            None
        """
        return self.swarm.state.update(other=other, inactives=inactives, **kwargs)

    def _prepare_tensors(self, **kwargs):
        if kwargs:
            step_data = kwargs
        else:
            step_data = self.get_input_data()
        return step_data

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ):
        """
        Reset the internal state of the :class:`SwarmComponent`.

        Args:
            inplace (bool, optional): Unused. Defaults to True.
            root_walker (Optional[StateData], optional): Set the internal state of
                the :class:`SwarmComponent` to this value. Defaults to None.
            states (Optional[StateData], optional): Set the internal state of the
                :class:`SwarmComponent` to this value. Defaults to None.
            kwargs: Other parameters required to reset the component.
        """
        pass


class EnvironmentAPI(SwarmComponent):
    """
    The Environment is in charge of stepping the walkers, acting as a state \
    transition function.

    For every different problem, a new Environment needs to be implemented
    following the :class:`EnvironmentAPI` interface.

    """

    default_inputs = {"actions": {}}
    default_outputs = ("observs", "rewards", "oobs")

    def __init__(
        self,
        action_shape: tuple,
        action_dtype,
        observs_shape: tuple,
        observs_dtype,
        swarm: "SwarmAPI" = None,
    ):
        """
        Initialize the :class:`EnvironmentAPI`.

        Args:
            action_shape (tuple): _description_
            action_dtype (_type_): _description_
            observs_shape (_type_): _description_
            observs_dtype (_type_): _description_
            swarm (SwarmAPI, optional): _description_. Defaults to None.
        """
        self._action_shape = action_shape
        self._action_dtype = action_dtype
        self._observs_shape = observs_shape
        self._observs_dtype = observs_dtype
        super(EnvironmentAPI, self).__init__(
            swarm=swarm,
            param_dict=self.param_dict,
        )

    @property
    def observs_shape(self) -> tuple:
        """Return the shape of the observations."""
        return self._observs_dtype

    @property
    def observs_dtype(self):
        """Return the dtype of the observations."""
        return self._observs_dtype

    @property
    def action_shape(self) -> tuple:
        """Return the shape of the actions."""
        return self._action_shape

    @property
    def action_dtype(self):
        """Return the dtype of the actions."""
        return self._action_dtype

    @property
    def param_dict(self) -> StateDict:
        """Return the dictionary defining all the data attributes that the component requires."""
        param_dict = {
            "observs": {"shape": self._observs_shape, "dtype": self._observs_dtype},
            "rewards": {"dtype": dtype.float32},
            "oobs": {"dtype": dtype.bool},
            "actions": {"shape": self._action_shape, "dtype": self._action_dtype},
        }
        return param_dict

    def step(self, **kwargs) -> StateData:
        """
        Return the data corresponding to the new state of the environment after \
        using the input data to make the corresponding state transition.
        """
        raise NotImplementedError()

    def make_transitions(
        self,
        inplace: bool = True,
        inactives: bool = False,
        **kwargs,
    ) -> Union[None, StateData]:
        """
        Return the data corresponding to the new state of the environment after \
        using the input data to make the corresponding state transition.

        Args:
            inplace: If ``False`` return the new data. If ``True``, update the state of the Swarm.
            inactives: Whether to update the walkers marked as inactive.
            **kwargs: Keyword arguments passed if the returned value from the
                ``states_to_data`` function of the class was a dictionary.

        Returns:
            Dictionary containing the data representing the state of the environment
            after the state transition. The keys of the dictionary are the names of
            the data attributes and its values are arrays representing a batch of
            new values for that attribute.

            The :class:`StatesEnv` returned by ``step`` will contain the returned
            data.
        """
        input_data = self._prepare_tensors(**kwargs)
        out_data = self.step(**input_data)
        if inplace:
            self.update(**out_data, inactives=inactives)
        return out_data

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        inactives: bool = True,
        **kwargs,
    ) -> Union[None, StateData]:
        """
        Reset the internal state of the :class:`SwarmComponent`.

        Args:
            inplace (bool, optional): If ``False`` return the new data. If ``True``,
                update the state of the Swarm. Defaults to ``True``.
            root_walker (Optional[StateData], optional): _description_. Defaults to None.
            states (Optional[StateData], optional): _description_. Defaults to None.
            inactives (bool, optional): Whether to update the walkers marked as inactive.
                Defaults to ``True``.
            kwargs: Other arguments passed to make_transitions.

        Returns:
            Union[None, StateData]: Dictionary containing the data representing
            the state of the environment after the state transition. The keys of
            the dictionary are the names of the data attributes and its values
            are arrays representing a batch of new values for that attribute.

            The :class:`StatesEnv` returned by ``step`` will contain the returned
            data.
        """
        return self.make_transitions(inplace=inplace, inactives=inactives, **kwargs)


class PolicyAPI(SwarmComponent):
    """
    The policy is in charge of calculating the interactions with the :class:`Environment`.

    The PolicyAPI class is responsible for defining the policy that determines the actions
    for interacting with the :class:`Environment` in a swarm simulation. This is an abstract
    base class, and specific policy implementations should inherit from this class
    and implement the 'select_actions' method.
    """

    default_outputs = tuple(["actions"])

    def select_actions(self, **kwargs) -> Union[Tensor, StateData]:
        """
        Select actions for each walker in the swarm based on the current state.

        This method must be implemented by subclasses.

        Args:
            **kwargs: Additional keyword arguments required for selecting actions.

        Returns:
            Union[Tensor, StateData]: The selected actions as a Tensor or a StateData dictionary.
        """
        raise NotImplementedError

    def act(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        """
        Calculate SwarmState containing the data needed to interact with the environment.

        Args:
            inplace (bool, optional): If True, updates the swarm state with the selected actions.
                                      If False, returns the selected actions. Defaults to True.
            **kwargs: Additional keyword arguments required for selecting actions.

        Returns:
            Union[None, StateData]: None if inplace is True. Otherwise, a StateData dictionary
                                    containing the selected actions.
        """
        action_input = self._prepare_tensors(**kwargs)
        actions_data = self.select_actions(**action_input)
        if not isinstance(actions_data, dict):
            actions_data = {"actions": actions_data}
        if inplace:
            self.update(**actions_data)
        else:
            return actions_data

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ) -> Union[None, StateData]:
        """
        Reset the internal state of the :class:`PolicyAPI`.

        Args:
            inplace (bool, optional): If True, updates the swarm state with the selected actions.
                                      If False, returns the selected actions. Defaults to True.
            root_walker (Optional[StateData], optional): Set the internal state of
                the PolicyAPI to this value. Defaults to None.
            states (Optional[StateData], optional): Set the internal state of the
                PolicyAPI to this value. Defaults to None.
            **kwargs: Other parameters required to reset the component.

        Returns:
            Union[None, StateData]: None if inplace is True. Otherwise, a StateData dictionary
                                    containing the selected actions.
        """
        # TODO: only run act when inputs are not present in root_walker/states
        # if root_walker is None and states is None:
        return self.act(inplace=inplace, **kwargs)


class Callback(SwarmComponent):
    """
    The :class:`Walkers` is a data structure that takes care of all the data involved \
    in making a Swarm evolve.
    """

    name = None

    def before_reset(self):
        pass

    def after_reset(self):
        pass

    def before_evolve(self):
        pass

    def after_evolve(self):
        pass

    def before_policy(self):
        pass

    def after_policy(self):
        pass

    def before_env(self):
        pass

    def after_env(self):
        pass

    def before_walkers(self):
        pass

    def after_walkers(self):
        pass

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ):
        pass

    def evolution_end(self) -> bool:
        return False

    def run_end(self):
        pass


class WalkersMetric(SwarmComponent):
    def __call__(self, inplace: bool = True, **kwargs) -> Tensor:
        input_data = self._prepare_tensors(**kwargs)
        out_data = self.calculate(**input_data)
        if inplace:
            self.update(**out_data)
        return out_data

    def calculate(self, **kwargs):
        raise NotImplementedError()

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ):
        pass


class WalkersAPI(SwarmComponent):
    """
    The WalkersAPI class defines the base functionality for managing walkers in a swarm simulation.
    This class inherits from the SwarmComponent class.
    """

    def get_input_data(self) -> StateData:
        """
        Return a dictionary with all the data that this component requires as inputs, including
        the data for inactive walkers.

        Returns:
            StateData: A dictionary containing all the required data for the WalkersAPI component,
                       where each key is the name of a data attribute and its value is a
                       multi-dimensional array.
        """

        def get_one_input(name, values):
            return self.get(
                name=name,
                default=values.get("default"),
                raise_error=not values.get("optional", False),
                inactives=True,
            )

        return {k: get_one_input(k, v) for k, v in self.inputs.items()}

    def update(
        self,
        other: Union["SwarmState", Dict[str, Tensor]] = None,
        inactives: bool = True,
        **kwargs,
    ) -> None:
        return super(WalkersAPI, self).update(other=other, inactives=inactives, **kwargs)

    def balance(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        """
        Perform a balance operation on the swarm state.

        Args:
            inplace (bool, optional): If True, updates the swarm state with the balanced data.
                                      If False, returns the balanced data. Defaults to True.
            **kwargs: Additional keyword arguments required for running the balance operation.

        Returns:
            Union[None, StateData]: None if inplace is True. Otherwise, a StateData dictionary
                                    containing the balanced data.
        """
        input_data = self._prepare_tensors(**kwargs)
        out_data = self.run_epoch(inplace=inplace, **input_data)
        if inplace:
            self.update(**out_data)
        return out_data

    def run_epoch(self, inplace: bool = True, **kwargs) -> StateData:
        """
        Implement the functionality for running an epoch in the derived class. This method is \
        called during the balance operation.

        Args:
            inplace (bool, optional): If `True`, updates the swarm state with the data generated
                during the epoch. If False, returns the data. Defaults to `True`.
            **kwargs: Additional keyword arguments required for running the epoch.

        Returns:
            StateData: A dictionary containing the data generated during the epoch if inplace is
                       False. Otherwise, returns None.
        """
        raise NotImplementedError()

    def reset(self, inplace: bool = True, **kwargs):
        """
        Reset the internal state of the Walkers class. This method should be implemented
        in the derived class.

        Args:
            inplace (bool, optional): Unused. Defaults to True.
            **kwargs: Additional keyword arguments required for resetting the component.

        Returns:
            None
        """
        pass

    def get_in_bounds_compas(self, oobs=None) -> Tensor:
        """
        Return an array of indexes corresponding to alive walkers chosen at random.

        Args:
            oobs (Optional[np.ndarray], optional): An optional boolean array indicating
                out-of-bounds walkers. If not provided, all walkers are considered alive.
                Defaults to None.

        Returns:
            Tensor: An array of indexes corresponding to randomly chosen alive walkers.
        """
        n_walkers = len(oobs) if oobs is not None else self.swarm.n_walkers
        indexes = judo.arange(n_walkers, dtype=int)
        # No need to sample if all walkers are dead or terminal.
        if oobs is None or oobs.all():
            return indexes
        alive_indexes = indexes[~oobs]
        compas_ix = random_state.permutation(alive_indexes)
        compas = random_state.choice(compas_ix, len(oobs), replace=True)
        compas[: len(compas_ix)] = compas_ix
        return compas

    def clone_walkers(self, will_clone=None, compas_clone=None, **kwargs) -> None:
        """
        Sample the clone probability distribution and clone the walkers accordingly.

        Args:
            will_clone (Optional[np.ndarray], optional): A boolean array indicating which walkers
                will be cloned. If not provided, the clone operation will be performed on all
                walkers. Defaults to None.
            compas_clone (Optional[np.ndarray], optional): An array of indexes indicating the
                walkers to be cloned. If not provided, random alive walkers will be chosen.
                Defaults to None.
            **kwargs: Additional keyword arguments required for cloning the walkers.

        Returns:
            None
        """
        self.swarm.state.clone(
            will_clone=will_clone,
            compas_clone=compas_clone,
            clone_names=self.swarm.clone_names,
        )


class SwarmAPI:
    """
    The Swarm implements the iteration logic to make the :class:`Walkers` evolve.

    It contains the necessary logic to use an Environment, a Model, and a \
    Walkers instance to create the algorithm execution loop.

    This class defines a method called run() that receives two optional arguments, \
    root_walker and state, and has no return value. This method runs the fractal AI Swarm \
    evolution process until a stop condition is met.

    In its implementation, it calls several other methods: \
    (before_reset(), reset(), after_reset(), evolve(), before_evolve(), after_evolve(), \
    evolution_end(), before_env(), after_env(), before_policy(), after_policy(), \
    before_walkers(), after_walkers(), run_end()) defined within the same class, \
    which are mainly used to manage different aspects of the search process or to \
    invoke user-defined callbacks.

    The evolve() method updates the states of the search environment and model, \
    makes the walkers undergo a perturbation process, and balances them. It also \
    invokes several other methods that trigger user-defined callbacks.

    The evolution_end() method returns ``True`` if any of the following conditions is met:
    1. The current epoch exceeds the maximum allowed epochs.
    2. All walkers are out of the problem domain.
    3. Any callback of the class has set the ``evolution_end`` flag to ``True``.

    Attributes:
        walkers_last (bool): If `True` indicates that the :class:`~fragile.Walkers` \
        class runs after acting on the environment. If `Fase`, the walkers run before \
        acting on the environment.

    Args:
        n_walkers (int): The number of walkers in the swarm.
        env (:class:`EnvironmentAPI`): An environment that simulates the objective function.
        policy (:class:`PolicyAPI`): A policy that defines how the individuals evolve.
        walkers (:class:`WalkersAPI`): A set of motion rules to control a population of walkers.
        callbacks (Optional[Iterable[Callback]]): A list of functions to call at each iteration.
        minimize (bool): If ``True``, take the minimum value of fitness, else take the maximum.
        max_epochs (int): Maximum number of epochs allowed before the swarm search is stopped.
    """

    walkers_last = True

    def __init__(
        self,
        n_walkers: int,
        env: EnvironmentAPI,
        policy: PolicyAPI,
        walkers: WalkersAPI,
        callbacks: Optional[Iterable[Callback]] = None,
        minimize: bool = False,
        max_epochs: int = 1e100,
    ):
        """Initialize a :class:`SwarmAPI`."""
        self.minimize = minimize
        self.max_epochs = int(max_epochs)
        self._n_walkers = n_walkers
        self._epoch = 0
        self._env = env
        self._policy = policy
        self._walkers = walkers
        self._state = None
        self._inputs = {}
        self._clone_names = set()
        self._callbacks = {}
        callbacks = callbacks if callbacks is not None else []
        for callback in callbacks:
            self.register_callback(callback, setup=False)
        self.setup()

    @property
    def n_walkers(self) -> int:
        """Return the number of walkers in the swarm."""
        return self._n_walkers

    @property
    def n_actives(self) -> int:
        """Returns the number of active walkers in the swarm."""
        return self.state.n_actives

    @property
    def epoch(self) -> int:
        """Return the current epoch of the search algorithm."""
        return self._epoch

    @property
    def state(self) -> SwarmState:
        """Returns the state instance describing the walkers of the swarm."""
        return self._state

    @property
    def env(self) -> EnvironmentAPI:
        """All the simulation code (problem specific) will be handled here."""
        return self._env

    @property
    def policy(self) -> PolicyAPI:
        """
        All the policy and random perturbation code (problem specific) will \
        be handled here.
        """
        return self._policy

    @property
    def walkers(self) -> WalkersAPI:
        """Access the :class:`Walkers` in charge of implementing the FAI evolution process."""
        return self._walkers

    @property
    def callbacks(self) -> Dict[str, Callback]:
        """Return the dictionary containing all the user-defined callbacks."""
        return self._callbacks

    @property
    def param_dict(self) -> StateDict:
        """Return the copy of parameters dictionary describing the attributes of the walkers."""
        return copy.deepcopy(self.state.param_dict)

    @property
    def clone_names(self) -> Set[str]:
        """Return the set of all the attributes that are cloned when iterating the Swarm."""
        return self._clone_names

    @property
    def inputs(self) -> dict:
        """Return the dictionary containing all the inputs of the search algorithm."""
        return self._inputs

    def __len__(self) -> int:
        """Return the the number of walkers in the swarm."""
        return self.n_walkers

    def __getattr__(self, item):
        if item in self.callbacks:
            return self.callbacks[item]
        return super(SwarmAPI, self).__getattribute__(item)

    def setup_state(self, param_dict: StateDict, n_walkers: Optional[int] = None):
        """
        Set up :class:`SwarmState` instance for the search.

        Args:
            param_dict: Initial dictionary of parameters.
            n_walkers: Number of :class:`Walker` instances. If not set, uses the previous number.

        Returns:
            None.
        """
        if n_walkers is not None:
            self._n_walkers = n_walkers
        self._state = SwarmState(n_walkers=self.n_walkers, param_dict=param_dict)
        self._state.reset()

    def setup(self) -> None:
        """
        Prepare :class:`Swarm` and internal components.

        Returns:
            None.
        """
        self._setup_components()
        self._setup_clone_names()
        self._setup_inputs()

    def register_callback(self, callback: Callback, setup: bool = True) -> None:
        """
        Register a ``Callback`` object with the :class:`Swarm`.

        When `setup=True`, calls setup() on the callback. Adds the Callback's
        attribute dictionary to the instance's :class:`State`.

        This allows the state to manage all the data the callback needs.

        Args:
            callback: An instance of :class:`Callback`.
            setup: Indicates whether we should call `setup()` on the given ``callback``.

        Returns:
            None.
        """
        if setup:
            callback.setup(self)
            new_param_dict = {**self.param_dict, **callback.param_dict}
            self.setup_state(n_walkers=self.n_walkers, param_dict=new_param_dict)
        self.callbacks[callback.name] = callback
        clone_names = [k for k, v in callback.inputs.items() if v.get("clone")]
        self._clone_names = set(list(self.clone_names) + clone_names)

    def get(self, name: str, default: Any = None, raise_error: bool = False) -> Any:
        """
        Access attributes of the :class:`Swarm` and its children.

        Args:
            name: Name of the attribute
            default: Default value to return if attribute doesn't exist.
            raise_error: Raise an error if the named attribute cannot be found.

        Returns:
            The attribute if it exists, else the default.
        """
        return self.state.get(name=name, default=default, raise_error=raise_error)

    def reset(
        self,
        root_walker: Optional["OneWalker"] = None,
        state: Optional[SwarmState] = None,
    ) -> None:
        """
        Reset a :class:`~fragile.Swarm` and clear the internal data to start a \
        new search process.

        Upon being called, ``reset`` clears all internal data from previous
        runs and sets up the Swarm to begin a new search process.

        Args:
            root_walker (:class:`~fragile.OneWalker`, optional): A walker representing the initial
                state of the search. The walkers will be reset to thiswalker and it will be added
                to the root of the :class:`~fragile.StateTree`.
            state (:class:`~fragile.SwarmState`): Defines the initial state of the `Swarm`.
                Defaults are loaded from the current setup given during instantiation
                of the `Swarm()` object.
        """
        self.state.reset(root_walker=root_walker)
        if not self.walkers_last:
            self.walkers.reset(root_walker=root_walker)
        self.env.reset(root_walker=root_walker)
        self.policy.reset(root_walker=root_walker)
        if self.walkers_last:
            self.walkers.reset(root_walker=root_walker)
        for callback in self.callbacks.values():
            callback.reset(root_walker=root_walker)
        self._epoch = 0

    def run(
        self,
        root_walker: Optional[StateData] = None,
        state: Optional[StateData] = None,
    ) -> None:
        """
        Run a new search process until the stop condition is met.

        This method runs a new search process by resetting the walkers to the initial state given
        in the ``root_walker`` argument. If not provided, it will use the previously set \
        root walker. The initial swarm state can be defined using the `state` dictionary.

        During each epoch of the search process, the swarm undergoes an evolution process in the
        environment, and the search continues until the stop condition is met. The stop condition
        is defined by either of the following conditions:
            - Maximum number of epochs are exceeded (defined in ``max_epochs`` attribute).
            - All the walkers are out of bounds (defined in ``oobs`` attribute of the ``walkers``).
            - Any user-defined callback returns True for its ``evolution_end`` method.

        After the stop condition is met, the ``run_end`` method is called and all callbacks'
        ``run_end`` methods are executed.

        Args:
            root_walker (Optional[StateData]): Walker representing the initial state of the search.
             The walkers will be reset to this walker, and it will
             be added to the root of the :class:`StateTree` if any.
            state (Optional[StateData]): StateData dictionary that define the
             initial state of the Swarm.

        Returns:
            None.
        """
        self.before_reset()
        self.reset(root_walker=root_walker)
        self.after_reset()
        while not self.evolution_end():
            self.before_evolve()
            self.evolve()
            self.after_evolve()
        self.run_end()

    def evolution_end(self) -> bool:
        """
        Check if the :class:`Swarm`'s evolution process should stop.

        Checks whether the maximum number of epochs has been reached,
        all walkers are out of bounds or any of the callbacks have ended their evolution.

        Returns:
            bool: A boolean value indicating whether the evolution process has to stop.
        """
        oobs = self.get("oobs")
        return (
            (self.epoch >= self.max_epochs)
            or oobs.all()
            or any(c.evolution_end() for c in self.callbacks.values())
        )

    def evolve(self) -> None:
        """
        Make the walkers undergo a perturbation process in the swarm \
        :class:`Environment`.

        The evolution algorithm function undergoes the following steps:
         2. If the ``self.walkers_last`` is ``False`` then it calls the callback functions \
            declared as 'before_walkers'.
         3.  If the ``self.walkers_last`` is ``False`` It balances the walkers according \
            to the clone probabilities by calling the balance function on the \
            :class:`Walkers` instance that is a component of  the :class:`Swarm` class.
         4. If the ``self.walkers_last`` is ``False``  After the balancing is complete, \
            the function calls more callback functions declared as 'after_walkers'.
         5. It then calls more callback functions defined under the 'before_policy' tag.
         6. Makes each walker select an action by calling the act() method of the policy object.
         7. Performs transition based on the new state created after taking the given action.
         8. Calls more callback functions defined under the 'after_env' tag.
         9. If the ``self.walkers_last`` is ``True`` runs the walkers balance process and \
            related callbacks.
        This function updates the :class:`StatesEnv` and the :class:`StatesModel` after each step.
        """
        if not self.walkers_last:
            self.before_walkers()
            self.walkers.balance()
            self.after_walkers()
        self.before_policy()
        self.policy.act()
        self.after_policy()
        self.before_env()
        self.env.make_transitions()
        self.after_env()
        if self.walkers_last:
            self.before_walkers()
            self.walkers.balance()
            self.after_walkers()
        self._epoch += 1

    def before_reset(self) -> None:
        """ "Called before resetting the search process back to its initial state."""
        for callback in self.callbacks.values():
            callback.before_reset()

    def after_reset(self) -> None:
        """Called after resetting the search process back to its initial state."""
        for callback in self.callbacks.values():
            callback.after_reset()

    def before_policy(self) -> None:
        """Called before using the policy to sample an action."""
        for callback in self.callbacks.values():
            callback.before_policy()

    def after_policy(self) -> None:
        """Called after using the policy to sample an action."""
        for callback in self.callbacks.values():
            callback.after_policy()

    def before_env(self) -> None:
        """Called before the environment has been updated."""
        for callback in self.callbacks.values():
            callback.before_env()

    def after_env(self) -> None:
        """Called after the environment has been updated."""
        for callback in self.callbacks.values():
            callback.after_env()

    def before_walkers(self) -> None:
        """Called before the walkers have been balanced."""
        for callback in self.callbacks.values():
            callback.before_walkers()

    def after_walkers(self) -> None:
        """Called after the walkers have been balanced."""
        for callback in self.callbacks.values():
            callback.after_walkers()

    def before_evolve(self) -> None:
        """Called before the evolve step has started."""
        for callback in self.callbacks.values():
            callback.before_evolve()

    def after_evolve(self) -> None:
        """Called after the evolve step has ended."""
        for callback in self.callbacks.values():
            callback.after_evolve()

    def run_end(self) -> None:
        """Called after the evolution process has ended."""
        for callback in self.callbacks.values():
            callback.run_end()



from collections import defaultdict
from typing import Any, Dict, Optional, Tuple, Union

import judo
from judo import dtype, random_state, tensor

from fragile.core.api_classes import WalkersAPI, WalkersMetric
from fragile.core.fractalai import fai_iteration, relativize
from fragile.core.typing import InputDict, StateData, StateDict, Tensor


class SimpleWalkers(WalkersAPI):
    default_inputs = {
        "observs": {},
        "oobs": {"optional": True},
        "is_won": {"optional": True},
        "rewards": {},
        "scores": {"clone": True},
    }
    default_param_dict = {"scores": {"dtype": dtype.float32}}
    default_outputs = tuple(["scores"])

    def __init__(
        self,
        accumulate_reward: bool = True,
        score_scale: float = 1.0,
        diversity_scale: float = 1.0,
        minimize: bool = False,
        **kwargs,
    ):
        self.score_scale = score_scale
        self.diversity_scale = diversity_scale
        self.accumulate_reward = accumulate_reward
        self.minimize = minimize
        super(SimpleWalkers, self).__init__(**kwargs)

    def run_epoch(
        self,
        observs,
        rewards,
        scores,
        oobs=None,
        inplace: bool = True,
        **kwargs,
    ) -> StateData:
        scores = rewards + scores if self.accumulate_reward else rewards
        sign_scores = -1.0 * scores if self.minimize else scores
        compas_ix, will_clone = fai_iteration(
            observs=observs,
            rewards=sign_scores,
            oobs=oobs,
            dist_coef=self.diversity_scale,
            reward_coef=self.score_scale,
        )
        if inplace:
            self.clone_walkers(compas_clone=compas_ix, will_clone=will_clone)
        return {"scores": scores}


def l2_norm(x: Tensor, y: Tensor) -> Tensor:
    """
    Compute the L2 norm between two tensors.

    Args:
        x (Tensor): The first tensor.
        y (Tensor): The second tensor.

    Returns:
        Tensor: The L2 norm between the two input tensors.
    """
    return judo.sqrt(judo.sum((x - y) ** 2, 1))


class ScoreMetric(WalkersMetric):
    """
    A base class for score metrics in a swarm of walkers.

    Attributes:
        default_param_dict (dict): Default parameters dictionary for the score metric.
        default_outputs (tuple): Tuple containing the default output keys.
    """

    default_param_dict = {"scores": {"dtype": dtype.float32}}
    default_outputs = tuple(["scores"])


class RewardScore(ScoreMetric):
    """
    A class representing the reward score metric for a swarm of walkers.

    Attributes:
        default_inputs (dict): Default inputs dictionary for the reward score metric.
    """

    default_inputs = {"rewards": {}}

    def __init__(self, accumulate_reward: bool = True, keep_max_reward: bool = False, **kwargs):
        self.accumulate_reward = accumulate_reward
        self.keep_max_reward = keep_max_reward
        super(RewardScore, self).__init__(**kwargs)

    @property
    def inputs(self) -> InputDict:
        inputs = super(RewardScore, self).inputs
        if self.accumulate_reward:
            inputs["scores"] = {"clone": True}
        return inputs

    def calculate(
        self,
        rewards: Optional[Tensor] = None,
        scores=None,
        **kwargs,
    ) -> Dict[str, Tensor]:
        """
        Calculate the scores for the walkers based on rewards.

        Args:
            rewards (array): Array of walkers' rewards.
            scores (array, optional): Array of walkers' scores.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: A dictionary containing the scores.
        """
        rewards = self.get("rewards", inactives=True) if rewards is None else rewards
        actives = self.swarm.state.actives
        if self.accumulate_reward:
            values = self.get("scores", inactives=True)
            values[actives] = values[actives] + rewards[actives]
        else:
            values = rewards
        if self.keep_max_reward and scores is not None:
            values = (
                judo.minimum(values, scores)
                if self.swarm.minimize
                else judo.maximum(values, scores)
            )
        return {"scores": values}

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ) -> None:
        """
        Reset the reward score metric.

        Args:
            inplace (bool, optional): Whether to reset in-place. Defaults to True.
            root_walker (StateData, optional): The root walker to reset. Defaults to None.
            states (Optional[StateData], optional): The state data to reset. Defaults to None.
            **kwargs: Additional keyword arguments.
        """
        if root_walker is None and not self.accumulate_reward:
            rewards = self.get("rewards", inactives=True)
            # _default_reward = numpy.inf if self.swarm.minimize else numpy.ninf
            # rewards = rewards * _default_reward
            self.update(scores=rewards, inactives=True)


class SonicScore(ScoreMetric):
    accumulate_reward = False
    default_inputs = {
        "rewards": {"clone": True},
        "infos": {"clone": True},
        "scores": {"clone": True},
    }

    @staticmethod
    def score_from_info(info):
        not_in_bonus_level = not (info.get("in_bonus_level") or info.get("in_transition_screen"))
        if info.get("in_boss_fight", False):
            score_x = 1010
        elif info.get("in_bonus_level"):
            score_x = 1002
        elif info.get("in_transition_screen"):
            score_x = 1008
        else:
            score_x = 1000 * min(info.get("x", 0) / max(info.get("screen_x_end", 1), 1), 1)
        score = (
            info.get("score", 0)
            + score_x
            + 10 * info.get("rings", 0)
            + 1001
            * (
                info.get("act", 0)
                if (not_in_bonus_level and not info.get("in_transition_screen"))
                else info.get("act", 0) - 1
            )
            + 5000 * info.get("zone", 0)
            + 1000 * info.get("lives", 0)
        )
        return int(score)

    def calculate(self, rewards, scores=None, **kwargs):
        scores = tensor([self.score_from_info(info) for info in self.get("infos", inactives=True)])
        scores = judo.maximum(scores, self.get("scores", inactives=True))
        return {"scores": scores}


class MarioScore(ScoreMetric):
    name = "MarioScore"
    accumulate_reward = False
    default_inputs = {"rewards": {}, "infos": {}}

    @staticmethod
    def score_from_info(info):
        score = (
            (info.get("world", 0) * 25000)
            + (info.get("stage", 0) * 5000)
            + info.get("x_pos", 0)
            + 10 * int(bool(info.get("in_pipe", 0)))
            + 100 * int(bool(info.get("flag_get", 0)))
            + 10 * info.get("coins", 0)
            + info["life"] * 1000
            # + (abs(info["x_pos"] - info["x_position_last"]))
        )

        return score

    def calculate(self, rewards=None, scores=None, **kwargs):
        scores = tensor([self.score_from_info(info) for info in self.get("infos")])
        self.update(scores=scores)
        scores = self.get("scores", inactives=True)
        return {"scores": scores}


class DiversityMetric(WalkersMetric):
    """
    A base class for diversity metrics in a swarm of walkers.

    Attributes:
        default_param_dict (dict): Default parameters dictionary for the diversity metric.
        default_outputs (tuple): Tuple containing the default output keys.
    """

    default_param_dict = {"diversities": {"dtype": dtype.float32}}
    default_outputs = tuple(["diversities"])


class RandomDistance(DiversityMetric):
    """
    A class representing the random distance diversity metric for a swarm of walkers.

    Attributes:
        default_inputs (dict): Default inputs dictionary for the random distance diversity metric.
    """

    default_inputs = {"observs": {}, "oobs": {}}

    def calculate(self, observs, oobs, **kwargs):
        """
        Calculate the diversities for the walkers based on the L2 distance to another \
        walker chosen at random.

        Args:
            observs (array): Array of walkers' observations.
            oobs (array): Array of out-of-bounds flags for walkers.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: A dictionary containing the diversities.
        """
        compas = self.swarm.walkers.get_in_bounds_compas(oobs=oobs)
        obs = judo.astype(observs.reshape(observs.shape[0], -1), dtype.float32)
        if hasattr(self.swarm.env, "bounds"):
            deltas = self.swarm.env.bounds.pbc_distance(obs, obs[compas])
            # return {"diversities": numpy.linalg.norm(deltas, axis=1).flatten()}
            return {"diversities": l2_norm(deltas, 0).flatten()}
        return {"diversities": l2_norm(obs, obs[compas]).flatten()}


class Walkers(WalkersAPI):
    """
    Walkers class handles the walkers' dynamics, including scoring, diversity, cloning, \
    and resetting.

    Inherits from WalkersAPI.
    """

    default_param_dict = {
        "compas_clone": {"dtype": dtype.int64},
        "virtual_rewards": {"dtype": dtype.float32},
        "clone_probs": {"dtype": dtype.float32},
        "will_clone": {"dtype": dtype.bool},
        # "actives": {"dtype": dtype.bool},
    }
    default_outputs = (
        "compas_clone",
        "virtual_rewards",
        "clone_probs",
        "will_clone",
    )  # , "actives")

    default_inputs = {"oobs": {}, "terminals": {"optional": True}}

    def __init__(
        self,
        score: ScoreMetric = None,
        diversity: DiversityMetric = None,
        minimize: bool = False,
        score_scale: float = 1.0,
        diversity_scale: float = 1.0,
        track_data=None,
        accumulate_reward: bool = True,
        keep_max_reward: bool = False,
        clone_period: int = 1,
        freeze_walkers=True,
        **kwargs,
    ):
        """
        Initialize a Walkers object.

        Args:
            score (ScoreMetric, optional): Scoring metric. Defaults to None.
            diversity (DiversityMetric, optional): Diversity metric. Defaults to None.
            minimize (bool, optional): Whether to minimize the score. Defaults to False.
            score_scale (float, optional): Scaling factor for score. Defaults to 1.0.
            diversity_scale (float, optional): Scaling factor for diversity. Defaults to 1.0.
            track_data (iterable, optional): Data to track. Defaults to None.
            accumulate_reward (bool, optional): Whether to accumulate rewards. Defaults to True.
            keep_max_reward (bool, optional): Whether to keep maximum reward. Defaults to False.
            clone_period (int, optional): Clone period. Defaults to 1.
            freeze_walkers (bool, optional): Whether to freeze walkers. Defaults to True.
            **kwargs: Additional keyword arguments.
        """

        self.minimize = minimize
        self.score_scale = score_scale
        self.diversity_scale = diversity_scale
        self.clone_period = clone_period
        self.score = (
            score
            if score is not None
            else RewardScore(accumulate_reward=accumulate_reward, keep_max_reward=keep_max_reward)
        )
        self.accumulate_reward = self.score.accumulate_reward
        self.diversity = diversity if diversity is not None else RandomDistance()
        self.track_data = set(track_data) if track_data is not None else set()
        self.freeze_walkers = freeze_walkers
        super(WalkersAPI, self).__init__(**kwargs)

    @property
    def param_dict(self) -> StateDict:
        return {
            **super(WalkersAPI, self).param_dict,
            **self.diversity.param_dict,
            **self.score.param_dict,
        }

    @property
    def inputs(self) -> InputDict:
        return {**super(WalkersAPI, self).inputs, **self.diversity.inputs, **self.score.inputs}

    @property
    def outputs(self) -> Tuple[str, ...]:
        return super(WalkersAPI, self).outputs + self.score.outputs + self.diversity.outputs

    def setup(self, swarm):
        """
        Set up the Walkers object by initializing its swarm.

        Args:
            swarm: Swarm object to be initialized.
        """
        super(Walkers, self).setup(swarm)
        self.diversity.setup(swarm)
        self.score.setup(swarm)
        self.minimize = swarm.minimize

    def balance(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        """
        Balance the walkers.

        Args:
            inplace (bool, optional): Whether to perform the operation inplace. Defaults to True.
            **kwargs: Additional keyword arguments.

        Returns:
            Union[None, StateData]: None if inplace=True, otherwise a StateData object.
        """
        if self.swarm.epoch % self.clone_period == 0 or self.swarm.epoch == 0:
            return super(Walkers, self).balance(inplace=inplace, **kwargs)

    def run_epoch(self, inplace: bool = True, oobs=None, **kwargs) -> Dict[str, Tensor]:
        """
        Execute an epoch in the walkers' dynamics.

        Args:
            inplace (bool, optional): Whether to perform the operation inplace. Defaults to True.
            oobs: Out of bounds flags.
            **kwargs: Additional keyword arguments.

        Returns:
            StateData: State data after the epoch execution.
        """
        scores = self.score(**kwargs)
        diversities = self.diversity(oobs=oobs, **kwargs)
        virtual_rewards = self.calculate_virtual_reward(**{**scores, **diversities})
        clone_data = self.calculate_clones(oobs=oobs, **virtual_rewards)
        actives = clone_data.get("actives")
        if actives is not None and self.freeze_walkers:
            self.swarm.state.update_actives(actives)
        if inplace:
            self.clone_walkers(**clone_data)
        return {**scores, **diversities, **virtual_rewards, **clone_data}

    def calculate_virtual_reward(self, scores, diversities, **kwargs):
        """
        Calculate the virtual rewards for walkers based on scores and diversities.

        Args:
            scores (array): Array of walkers' scores.
            diversities (array): Array of walkers' diversities.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: A dictionary containing the virtual rewards.
        """
        scores = -1.0 * scores if self.minimize else scores
        norm_scores = relativize(scores)
        norm_diver = relativize(diversities)
        virtual_rewards = norm_scores**self.score_scale * norm_diver**self.diversity_scale
        return {"virtual_rewards": virtual_rewards}

    def calculate_clones(self, virtual_rewards, oobs=None) -> Dict[str, Tensor]:
        """
        Calculate the walkers that will clone and their target companions.

        Args:
            virtual_rewards (array): Array of walkers' virtual rewards.
            oobs (array, optional): Array of out-of-bounds flags for walkers. Defaults to None.

        Returns:
            dict: A dictionary containing clone probabilities, will_clone flags, \
            companion clones, and active flags.
        """
        n_walkers = len(virtual_rewards)
        all_virtual_rewards_are_equal = (virtual_rewards == virtual_rewards[0]).all()
        terminals = self.swarm.get("terminals")
        if all_virtual_rewards_are_equal:
            clone_probs = judo.zeros(n_walkers, dtype=dtype.float)
            compas_clone = judo.arange(n_walkers)
        else:
            compas_clone = self.get_in_bounds_compas(oobs)
            # This value can be negative!!
            clone_probs = (virtual_rewards[compas_clone] - virtual_rewards) / virtual_rewards
        prob_trigger = judo.abs(clone_probs) > random_state.random_sample(n_walkers)
        will_clone = judo.logical_and(prob_trigger, clone_probs > 0)
        actives = judo.logical_not(judo.logical_and(prob_trigger, clone_probs < 0))
        if oobs is not None:
            will_clone[oobs] = True  # Out of bounds walkers always clone
        if terminals is not None:
            will_clone[terminals] = False  # Terminal walkers do not clone
            actives[terminals] = False
        return dict(
            clone_probs=clone_probs,
            will_clone=will_clone,
            compas_clone=compas_clone,
            actives=actives,
        )

    def reset(self, inplace: bool = True, **kwargs) -> None:
        """
        Reset the Walkers object and its score and diversity components.

        Args:
            inplace (bool, optional): Whether to perform the operation inplace. Defaults to True.
            **kwargs: Additional keyword arguments.
        """
        super(Walkers, self).reset(inplace=inplace, **kwargs)
        self.score.reset(**kwargs)
        self.diversity.reset(**kwargs)


class ExplorationWalkers(Walkers):
    def __init__(self, exploration_scale: float = 1.0, **kwargs):
        super(ExplorationWalkers, self).__init__(**kwargs)
        self._explore_counts = defaultdict(0)
        self.exploration_scale = exploration_scale

    @property
    def explore_counts(self) -> Dict[Any, int]:
        return self._explore_counts

    def calculate_virtual_reward(self, scores, diversities, **kwargs):
        """Apply the virtual reward formula to account for all the different goal scores."""
        vr_dict = super(ExplorationWalkers, self).calculate_virtual_reward(
            scores,
            diversities,
            **kwargs,
        )
        explore_reward = relativize(self.get_explore_rewards())
        virtual_rewards = vr_dict["virtual_rewards"] * explore_reward**self.exploration_scale
        return {"virtual_rewards": virtual_rewards}

    def get_explore_rewards(self):
        coords_keys = self.get_coords_keys()
        for key in coords_keys:
            self._explore_counts[key] += 1
        return 1 / judo.tensor([self.explore_counts[k] for k in coords_keys])

    def get_coords_keys(self):
        keys = []
        for info in self.get("infos"):
            in_bonus_level = info.get("screen_x_end", 1) == 0 and info.get("x", -1) != 0
            act = info.get("act") if not in_bonus_level else 0.5
            key = (info.get("zone"), act, int(info.get("x", 0) / 100), int(info.get("y", 0) / 100))
            keys.append(key)
        return keys


class NoBalance(Walkers):
    """
    A class representing walkers with no balancing behavior in a swarm.

    Inherits from the Walkers class and modifies the properties and methods to disable balancing.
    """

    @property
    def param_dict(self) -> StateDict:
        return self.score.param_dict

    @property
    def inputs(self) -> InputDict:
        return self.score.inputs

    @property
    def outputs(self) -> Tuple[str, ...]:
        return self.score.outputs

    def run_epoch(self, inplace: bool = True, oobs=None, **kwargs) -> Dict[str, Tensor]:
        """
        Run an epoch for the walkers without balancing.

        Args:
            inplace (bool, optional): Whether to run the epoch in-place. Defaults to True.
            oobs (array, optional): Array of out-of-bounds walkers. Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: A dictionary containing the scores.
        """
        scores = self.score(**kwargs)
        return scores



from typing import Tuple, Union

import numpy
import pandas as pd
import panel

from fragile.dataviz.stream_plots import Curve, Div, PlotCallback, RGB, Table


class PlotRootWalker(PlotCallback):
    name = "plot_root"

    def __init__(self, image_shape=None, step_after_set_state: bool = False, **kwargs):
        super(PlotRootWalker, self).__init__(**kwargs)
        self._image_available = False
        self._image_shape = image_shape
        self.table = Table(title="Run Summary", height=50)
        self.curve = Curve(
            data_names=["epoch", "score"],
            title="Score",
            xlabel="Epoch",
            ylabel="Score",
        )
        # self.swarm_curve = Curve
        self.info = Div(title="Root info")
        self.image = None
        self._last_score = -numpy.inf
        self._step_after_set_state = step_after_set_state

    @property
    def root(self):
        return self.swarm.root

    @property
    def image_shape(self) -> Union[None, Tuple[int, ...]]:
        return self._image_shape

    def setup(self, swarm):
        super(PlotRootWalker, self).setup(swarm)
        self._image_available = (
            hasattr(self.swarm.env, "plangym_env") or "rgb" in self.swarm.param_dict
        )
        if self._image_available:
            first_img = (
                self.root.rgb
                if "rgb" in self.root.data
                else self.swarm.env.plangym_env.get_image().astype(numpy.uint8)
            )
            if self.image_shape is None:
                h, w, *_ = first_img.shape
                self._image_shape = (h, w)
            self.image = RGB(data=first_img, title="Root Image")
            self.image.opts(
                height=self.image_shape[0],
                width=self.image_shape[1],
                xticks=None,
                yticks=None,
                xlabel=None,
                ylabel=None,
            )

    def send(self):
        if hasattr(self.root, "info"):
            self.info.send(repr(self.root.info))
        current_score = self.root.score
        wc, actives = self.get("will_clone", inactives=True), self.swarm.state.actives
        step = ~wc & actives

        summary_table = pd.DataFrame(
            columns=["epoch", "best_score", "pct_oobs", "pct_clone", "pct_freeze", "pct_step"],
            data=[
                [
                    self.swarm.epoch,
                    current_score,
                    self.get("oobs").mean(),
                    wc.mean(),
                    1 - actives.mean(),
                    step.mean(),
                ],
            ],
        )
        self.table.send(summary_table)
        score_data = pd.DataFrame(
            columns=["epoch", "score"],
            data=[[self.swarm.epoch, current_score]],
        )
        self.curve.send(score_data)
        if self._image_available and current_score != self._last_score:
            if "rgb" in self.root.data:
                img_data = self.root.rgb
            else:
                img_data = self.image_from_state(self.root.state)
            self.image.send(img_data)
        self._last_score = float(current_score)

    def panel(self):
        summary = panel.Row(self.table.plot, self.curve.plot)
        if self._image_available:
            return panel.Column(summary, panel.Row(self.info.plot, self.image.plot))
        return summary

    def image_from_state(self, state):
        self.swarm.env.plangym_env.set_state(state)
        if self._step_after_set_state:
            old_frameskip = int(self.swarm.env.plangym_env.frameskip)
            self.swarm.env.plangym_env.frameskip = 1
            self.swarm.env.plangym_env.step(self.swarm.env.plangym_env.sample_action())
            self.swarm.env.plangym_env.frameskip = old_frameskip
        return self.swarm.env.plangym_env.get_image().astype(numpy.uint8)

    def reset(self, root_walker=None, state=None, **kwargs):
        self.curve.data_stream.clear()
        return super(PlotRootWalker, self).reset(root_walker=root_walker, state=state, **kwargs)


class PlotMario(PlotRootWalker):
    def __init__(self, **kwargs):
        super(PlotRootWalker, self).__init__(**kwargs)
        self._image_available = False
        self.run_table = Table(title="Run Summary", height=50)
        self.info_table = Table(title="State info", height=50, width=600)
        self.curve = Curve(
            data_names=["epoch", "score"],
            title="Score",
            xlabel="Epoch",
            ylabel="Score",
        )
        self.image = None
        self._last_score = -numpy.inf

    def send(self):
        current_score = self.root.score
        summary_table = pd.DataFrame(
            columns=["epoch", "best_score", "pct_oobs"],
            data=[[self.swarm.epoch, current_score, self.get("oobs").mean()]],
        )
        info_df = pd.DataFrame(
            columns=list(self.root.info.keys()),
            data=[list(self.root.info.values())],
        )
        self.info_table.send(info_df)
        self.run_table.send(summary_table)
        score_data = pd.DataFrame(
            columns=["epoch", "score"],
            data=[[self.swarm.epoch, current_score]],
        )
        self.curve.send(score_data)
        if self._image_available and (
            current_score != self._last_score or self.root.always_update
        ):
            if "rgb" in self.root.data:
                img_data = self.root.rgb
            else:
                img_data = self.image_from_state(self.root.state)
            self.image.send(img_data)
        self._last_score = float(current_score)

    def panel(self):
        summary = panel.Row(self.run_table.plot, self.info_table.plot)
        plot_row = (
            panel.Row(self.curve.plot, self.image.plot)
            if self._image_available
            else self.curve.plot
        )
        return panel.Column(summary, plot_row)

    def image_from_state(self, state):
        self.swarm.env.plangym_env.set_state(state)
        # self.swarm.env.plangym_env.step(0)
        return self.swarm.env.plangym_env.get_image().astype(numpy.uint8)



import judo
from judo.data_types import dtype

from fragile.core.api_classes import Callback
from fragile.core.typing import StateDict


class StoreInitAction(Callback):
    name = "store_init_action"
    default_inputs = {"init_actions": {"clone": True}}
    default_outputs = ("init_actions",)

    @property
    def param_dict(self) -> StateDict:
        return {"init_actions": dict(self.swarm.param_dict["actions"])}

    def before_env(self):
        if self.swarm.epoch == 0:
            self.update(init_actions=judo.copy(self.get("actions")))


class TrackWalkersId(Callback):
    default_inputs = {"id_walkers": {"clone": True}, "parent_ids": {"clone": True}}
    default_param_dict = {
        "id_walkers": {"dtype": dtype.hash_type},
        "parent_ids": {"dtype": dtype.hash_type},
    }

    def update_ids(self, inactives: bool = True):
        with judo.Backend.use_backend("numpy"):
            name = "states" if "states" in self.swarm.state.names else "observs"
            actives = judo.to_numpy(self.swarm.state.actives)
            new_ids_all = self.swarm.state.hash_batch(name)
            parent_ids = judo.copy(judo.to_numpy(self.get("parent_ids", inactives=inactives)))
            new_ids = judo.copy(judo.to_numpy(self.get("id_walkers", inactives=True)))
            parent_ids[actives] = judo.copy(new_ids[actives])
            new_ids[actives] = new_ids_all[actives]
        self.update(
            parent_ids=judo.to_backend(parent_ids),
            id_walkers=judo.to_backend(new_ids),
            inactives=inactives,
        )

    def after_env(self):
        self.update_ids()



from typing import Callable, Tuple

import judo
from judo import Backend, Bounds, tensor, typing
import numpy
import numpy as np
from scipy.optimize import Bounds as ScipyBounds, minimize

from fragile.core.env import Function
from fragile.core.fractalai import relativize
from fragile.core.typing import StateData, StateDict
from fragile.optimize.utils import GradFuncWrapper


class Minimizer:
    """Apply ``scipy.optimize.minimize`` to a :class:`Function`."""

    def __init__(self, function: Function, bounds=None, **kwargs):
        """
        Initialize a :class:`Minimizer`.

        Args:
            function: :class:`Function` that will be minimized.
            bounds: :class:`Bounds` defining the domain of the minimization \
                    process. If it is ``None`` the :class:`Function` :class:`Bounds` \
                    will be used.
            *args: Passed to ``scipy.optimize.minimize``.
            **kwargs: Passed to ``scipy.optimize.minimize``.

        """
        self.env = function
        self.function = function.function
        self.bounds = self.env.bounds if bounds is None else bounds
        self.kwargs = kwargs

    def minimize(self, x: typing.Tensor):
        """
        Apply ``scipy.optimize.minimize`` to a single point.

        Args:
            x: Array representing a single point of the function to be minimized.

        Returns:
            Optimization result object returned by ``scipy.optimize.minimize``.

        """

        def _optimize(_x):
            try:
                _x = _x.reshape((1,) + _x.shape)
                y = self.function(_x)
            except (ZeroDivisionError, RuntimeError):
                y = numpy.inf
            return y

        bounds = ScipyBounds(
            ub=judo.to_numpy(self.bounds.high) if self.bounds is not None else None,
            lb=judo.to_numpy(self.bounds.low) if self.bounds is not None else None,
        )
        return minimize(_optimize, x, bounds=bounds, **self.kwargs)

    def minimize_point(self, x: typing.Tensor) -> Tuple[typing.Tensor, typing.Scalar]:
        """
        Minimize the target function passing one starting point.

        Args:
            x: Array representing a single point of the function to be minimized.

        Returns:
            Tuple containing a numpy array representing the best solution found, \
            and the numerical value of the function at that point.

        """
        optim_result = self.minimize(x)
        point = tensor(optim_result["x"])
        reward = tensor(float(optim_result["fun"]))
        return point, reward

    def minimize_batch(self, x: typing.Tensor) -> Tuple[typing.Tensor, typing.Tensor]:
        """
        Minimize a batch of points.

        Args:
            x: Array representing a batch of points to be optimized, stacked \
               across the first dimension.

        Returns:
            Tuple of arrays containing the local optimum found for each point, \
            and an array with the values assigned to each of the points found.

        """
        x = judo.to_numpy(judo.copy(x))
        with Backend.use_backend("numpy"):
            result = judo.zeros_like(x)
            rewards = judo.zeros((x.shape[0], 1))
            for i in range(x.shape[0]):
                new_x, reward = self.minimize_point(x[i, :])
                result[i, :] = new_x
                rewards[i, :] = float(reward)
        self.bounds.high = tensor(self.bounds.high)
        self.bounds.low = tensor(self.bounds.low)
        result, rewards = tensor(result), tensor(rewards)
        return result, rewards


class MinimizerWrapper(Function):
    """
    Wrapper that applies a local minimization process to the observations \
    returned by a :class:`Function`.
    """

    def __init__(self, function: Function, **kwargs):
        """
        Initialize a :class:`MinimizerWrapper`.

        Args:
            function: :class:`Function` to be minimized after each step.
            *args: Passed to the internal :class:`Optimizer`.
            **kwargs: Passed to the internal :class:`Optimizer`.

        """
        self.unwrapped = function
        self.minimizer = Minimizer(function=self.unwrapped, **kwargs)

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return the shape of the wrapped environment."""
        return self.unwrapped.shape

    @property
    def function(self) -> Callable:
        """Return the function of the wrapped environment."""
        return self.unwrapped.function

    @property
    def bounds(self) -> Bounds:
        """Return the bounds of the wrapped environment."""
        return self.unwrapped.bounds

    @property
    def custom_domain_check(self) -> Callable:
        """Return the custom_domain_check of the wrapped environment."""
        return self.unwrapped.custom_domain_check

    def __getattr__(self, item):
        return getattr(self.unwrapped, item)

    def __repr__(self):
        return self.unwrapped.__repr__()

    def setup(self, swarm):
        self.unwrapped.setup(swarm)

    def make_transitions(self, inplace: bool = True, **kwargs):
        """
        Perform a local optimization process to the observations returned after \
        calling ``make_transitions`` on the wrapped :class:`Function`.
        """
        function_data = self.unwrapped.make_transitions(inplace=False, **kwargs)
        new_points, rewards = self.minimizer.minimize_batch(function_data.get("observs"))
        # new_points, rewards = tensor(new_points), tensor(rewards)
        oobs = self.calculate_oobs(new_points, rewards)
        new_data = dict(observs=new_points, rewards=rewards.flatten(), oobs=oobs)
        if inplace:
            self.update(**new_data)
        else:
            return new_data


class ParticleSimulation(Function):
    default_inputs = {
        "actions": {},
        "observs": {"clone": True},
        "velocities": {"clone": True},
        "oobs": {"optional": True},
        "time": {"clone": True},
    }
    default_outputs = "observs", "rewards", "oobs", "velocities"

    def __init__(self, step_size=1.0, steps=10, **kwargs):
        self.step_size = step_size
        self.steps = steps
        super(ParticleSimulation, self).__init__(**kwargs)
        # self.grad_func = GradFuncWrapper(lambda theta: self.function(numpy.array([theta]))[0])
        self.grad_func = GradFuncWrapper(self.function)

    @classmethod
    def from_instance(
        cls,
        function: Function,
        step_size: float = 1,
        steps=1,
        x0=None,
        start_same_pos: bool = False,
    ):
        return cls(
            function=function.function,
            bounds=function.bounds,
            custom_domain_check=function.custom_domain_check,
            actions_as_perturbations=True,
            step_size=step_size,
            steps=steps,
            x0=x0,
            start_same_pos=start_same_pos,
        )

    @property
    def param_dict(self) -> StateDict:
        param_dict = super(ParticleSimulation, self).param_dict
        param_dict["velocities"] = dict(param_dict["observs"])
        param_dict["kinetic_energy"] = dict(param_dict["rewards"])
        param_dict["total_energy"] = dict(param_dict["rewards"])
        param_dict["time"] = dict(param_dict["rewards"])
        return param_dict

    def calculate_forces(self, observs, **kwargs):

        compas = observs[np.random.permutation(np.arange(len(observs)))]
        deltas = self.bounds.pbc_distance(observs, compas)
        distances = np.linalg.norm(deltas, axis=1)
        vectors = (deltas) / np.clip(distances.reshape(-1, 1), 1e-4, np.inf)
        r2 = relativize(distances)
        r2i = 1.0 / (r2**2)
        r6i = r2i * r2i * r2i
        epot = r6i * (r6i - 1.0) * 4
        forces = epot.reshape(-1, 1) * vectors
        return forces

    def fai_leapfrog(self, init_pos, init_momentum, grad, step_size, function, force=None):
        """Perfom a leapfrog jump in the Hamiltonian space
        INPUTS
        ------
        init_pos: ndarray[float, ndim=1]
            initial parameter position
        init_momentum: ndarray[float, ndim=1]
            initial momentum
        grad: float
            initial gradient value
        step_size: float
            step size
        function: callable
            it should return the log probability and gradient evaluated at theta
            logp, grad = function(theta)
        OUTPUTS
        -------
        new_position: ndarray[float, ndim=1]
            new parameter position
        new_momentum: ndarray[float, ndim=1]
            new momentum
        gradprime: float
            new gradient
        new_potential: float
            new lnp
        """
        potential, grad = function(init_pos)
        lj_force = 0  # self.calculate_forces(init_pos)
        pot = relativize(potential).reshape(-1, 1)
        grad = (grad / numpy.linalg.norm(grad, axis=1).reshape(-1, 1)) * pot
        # make half step in init_momentum
        new_momentum = init_momentum + 0.5 * step_size * (lj_force - grad + force)
        # make new step in theta
        new_position = init_pos + step_size * new_momentum
        new_position = self.bounds.pbc(new_position)
        # compute new gradient
        new_potential, gradprime = function(new_position)
        new_pot = relativize(new_potential).reshape(-1, 1)
        gradprime = (gradprime / numpy.linalg.norm(gradprime, axis=1).reshape(-1, 1)) * new_pot

        # make half step in init_momentum again
        new_force = 0  # self.calculate_forces(new_position)
        new_momentum = new_momentum + 0.5 * step_size * (new_force - gradprime + force)
        return new_position, new_momentum, gradprime, new_potential

    def step(self, actions, observs, velocities, **kwargs) -> StateData:
        """

        Sum the target action to the observations to obtain the new points, and \
        evaluate the reward and boundary conditions.

        Returns:
            Dictionary containing the information of the new points evaluated.

             ``{"states": new_points, "observs": new_points, "rewards": typing.Scalar array, \
             "oobs": boolean array}``

        """
        for _ in range(self.steps):
            _, grad = self.grad_func(observs)
            data = self.fai_leapfrog(
                init_pos=observs,
                init_momentum=velocities,
                grad=grad,
                step_size=self.step_size,
                function=self.grad_func,
                force=0,
            )
            new_position, velocities, gradprime, new_potential = data
            observs = self.bounds.pbc(judo.tensor(new_position))
        return dict(
            observs=observs,
            velocities=judo.tensor(velocities),
            rewards=judo.tensor(new_potential),
            oobs=judo.zeros(len(new_potential), dtype=judo.bool),
        )



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
#
import os
import sys
from unittest.mock import MagicMock


class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()


sys.path.insert(0, os.path.abspath("../../"))
sys.setrecursionlimit(1500)
MOCK_MODULES = ["pandas", "holoviews.streams", "holoviews", "PIL", "numba", "pyarrow"]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)
# -- Project information -----------------------------------------------------

project = "Fragile"
copyright = "2020, Fragile Technologies"
author = "Guillem Duran, Sergio Hernandez"

# The short X.Y version
from fragile import __version__


version = __version__
# The full version, including alpha/beta/rc tags
release = __version__

# -- General configuration ---------------------------------------------------
# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "**.ipynb_checkpoints"]
# The master toctree document.
master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # "sphinx.ext.autodoc",
    "autoapi.extension",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.imgmath",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc.typehints",
    "sphinx_book_theme",
    "myst_nb",
    "sphinxcontrib.mermaid",
    "sphinx.ext.githubpages",
    "sphinxcontrib.bibtex"
    # "m2r",
]

suppress_warnings = ["image.nonlocal_uri"]
autodoc_typehints = "description"
# Autoapi settings
autoapi_type = "python"
autoapi_dirs = ["../../fragile"]
autoapi_add_toctree_entry = True
# Make use of custom templates
autoapi_template_dir = "_autoapi_templates"
exclude_patterns.append("_autoapi_templates/index.rst")

# Ignore sphinx-autoapi warnings on multiple target description
suppress_warnings.append("ref.python")
# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = [".rst", ".md"]
# source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- BibTeX options ----------------------------------------------------------

bibtex_bibfiles = ["markdown/few_shot_learning/references.bib"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_title = ""
html_theme = "sphinx_book_theme"
html_logo = "_static/fragile_logo.png"
html_favicon = "_static/fragile_favicon.png"
html_theme_options = {
    "github_url": "https://github.com/fragiletech/fragile",
    "repository_url": "https://github.com/fragiletech/fragile",
    "repository_branch": "gh-pages",
    "home_page_in_toc": True,
    "path_to_docs": "docs",
    "show_navbar_depth": 1,
    "use_edit_page_button": True,
    "use_repository_button": True,
    "use_download_button": True,
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "notebook_interface": "classic",
    },
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# myst_parser options
nb_execution_mode = "off"
myst_heading_anchors = 2
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "substitution",
]

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "Fragiledoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    "papersize": "letterpaper",
    # The font size ('10pt', '11pt' or '12pt').
    #
    "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    #
    "preamble": "",
    # Latex figure (float) alignment
    #
    "figure_align": "htbp",
}


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "Fragile.tex",
        "Fragile Documentation",
        "Guillem Duran, Sergio Hernandez",
        "manual",
    )
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "fragile", "Fragile Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "Fragile",
        "Fragile Documentation",
        author,
        "Fragile",
        "One line description of project.",
        "Miscellaneous",
    )
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"https://docs.python.org/": None}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



import math
from typing import Callable

import judo
from judo import random_state, tensor
from judo.data_structures.bounds import Bounds
from numba import jit
import numpy as np

from fragile.core.env import Function


"""
This file includes several test functions for optimization described here:
https://en.wikipedia.org/wiki/Test_functions_for_optimization
"""


def sphere(x: np.ndarray) -> np.ndarray:
    return judo.sum(x**2, 1).flatten()


def rastrigin(x: np.ndarray) -> np.ndarray:
    dims = x.shape[1]
    A = 10
    result = A * dims + judo.sum(x**2 - A * judo.cos(2 * math.pi * x), 1)
    return result.flatten()


def eggholder(x: np.ndarray) -> np.ndarray:
    x, y = x[:, 0], x[:, 1]
    first_root = judo.sqrt(judo.abs(x / 2.0 + (y + 47)))
    second_root = judo.sqrt(judo.abs(x - (y + 47)))
    result = -1 * (y + 47) * judo.sin(first_root) - x * judo.sin(second_root)
    return result


def styblinski_tang(x) -> np.ndarray:
    return judo.sum(x**4 - 16 * x**2 + 5 * x, 1) / 2.0


def rosenbrock(x) -> np.ndarray:
    return 100 * judo.sum((x[:, :-2] ** 2 - x[:, 1:-1]) ** 2, 1) + judo.sum(
        (x[:, :-2] - 1) ** 2,
        1,
    )


def easom(x) -> np.ndarray:
    exp_term = (x[:, 0] - np.pi) ** 2 + (x[:, 1] - np.pi) ** 2
    return -judo.cos(x[:, 0]) * judo.cos(x[:, 1]) * judo.exp(-exp_term)


def holder_table(_x) -> np.ndarray:
    x, y = _x[:, 0], _x[:, 1]
    exp = np.abs(1 - (np.sqrt(x * x + y * y) / np.pi))
    return -np.abs(np.sin(x) * np.cos(y) * np.exp(exp))


@jit(nopython=True)
def _lennard_fast(state):
    state = state.reshape(-1, 3)
    npart = len(state)
    epot = 0.0
    for i in range(npart):
        for j in range(npart):
            if i > j:
                r2 = np.sum((state[j, :] - state[i, :]) ** 2)
                r2i = 1.0 / r2
                r6i = r2i * r2i * r2i
                epot = epot + r6i * (r6i - 1.0)
    epot = epot * 4
    return epot


def lennard_jones(x: np.ndarray) -> np.ndarray:
    result = np.zeros(x.shape[0])
    x = judo.to_numpy(x)
    assert isinstance(x, np.ndarray)
    for i in range(x.shape[0]):
        try:
            result[i] = _lennard_fast(x[i])
        except ZeroDivisionError:
            result[i] = np.inf
    result = judo.as_tensor(result)
    return result


class OptimBenchmark(Function):

    benchmark = None
    best_state = None

    def __init__(self, dims: int, function: Callable, **kwargs):
        bounds = self.get_bounds(dims=dims)
        super(OptimBenchmark, self).__init__(bounds=bounds, function=function, **kwargs)

    @staticmethod
    def get_bounds(dims: int) -> Bounds:
        raise NotImplementedError


class Sphere(OptimBenchmark):
    benchmark = tensor(0.0)

    def __init__(self, dims: int, **kwargs):
        super(Sphere, self).__init__(dims=dims, function=sphere, **kwargs)

    @staticmethod
    def get_bounds(dims):
        bounds = [(-1000, 1000) for _ in range(dims)]
        return Bounds.from_tuples(bounds)

    @property
    def best_state(self):
        return judo.zeros(self.shape)


class Rastrigin(OptimBenchmark):
    benchmark = tensor(0.0)

    def __init__(self, dims: int, **kwargs):
        super(Rastrigin, self).__init__(dims=dims, function=rastrigin, **kwargs)

    @staticmethod
    def get_bounds(dims):
        bounds = [(-5.12, 5.12) for _ in range(dims)]
        return Bounds.from_tuples(bounds)

    @property
    def best_state(self):
        return judo.zeros(self.shape)


class EggHolder(OptimBenchmark):
    benchmark = tensor(-959.64066271)

    def __init__(self, dims: int = None, **kwargs):
        super(EggHolder, self).__init__(dims=2, function=eggholder, **kwargs)

    @staticmethod
    def get_bounds(dims=None):
        bounds = [(-512.0, 512.0), (-512.0, 512.0)]
        # bounds = [(1, 512.0), (1, 512.0)]
        return Bounds.from_tuples(bounds)

    @property
    def best_state(self):
        return tensor([512.0, 404.2319])


class StyblinskiTang(OptimBenchmark):
    def __init__(self, dims: int, **kwargs):
        super(StyblinskiTang, self).__init__(dims=dims, function=styblinski_tang, **kwargs)

    @staticmethod
    def get_bounds(dims):
        bounds = [(-5.0, 5.0) for _ in range(dims)]
        return Bounds.from_tuples(bounds)

    @property
    def best_state(self):
        return judo.ones(self.shape) * -2.903534

    @property
    def benchmark(self):
        return tensor(-39.16617 * self.shape[0])


class Rosenbrock(OptimBenchmark):
    def __init__(self, dims: int, **kwargs):
        super(Rosenbrock, self).__init__(dims=dims, function=rosenbrock, **kwargs)

    @staticmethod
    def get_bounds(dims):
        bounds = [(-10.0, 10.0) for _ in range(dims)]
        return Bounds.from_tuples(bounds)

    @property
    def best_state(self):
        return judo.ones(self.shape)

    @property
    def benchmark(self):
        return tensor(0.0)


class Easom(OptimBenchmark):
    def __init__(self, dims: int = None, **kwargs):
        super(Easom, self).__init__(dims=2, function=easom, **kwargs)

    @staticmethod
    def get_bounds(dims):
        bounds = [(-100.0, 100.0) for _ in range(dims)]
        return Bounds.from_tuples(bounds)

    @property
    def best_state(self):
        return judo.ones(self.shape) * np.pi

    @property
    def benchmark(self):
        return tensor(-1)


class HolderTable(OptimBenchmark):
    def __init__(self, dims: int = None, *args, **kwargs):
        super(HolderTable, self).__init__(dims=2, function=holder_table, *args, **kwargs)

    @staticmethod
    def get_bounds(dims):
        bounds = [(-10.0, 10.0) for _ in range(dims)]
        return Bounds.from_tuples(bounds)

    @property
    def best_state(self):
        return tensor([8.05502, 9.66459])

    @property
    def benchmark(self):
        return tensor(-19.2085)


class LennardJones(OptimBenchmark):
    # http://doye.chem.ox.ac.uk/jon/structures/LJ/tables.150.html
    minima = {
        "2": -1,
        "3": -3,
        "4": -6,
        "5": -9.103852,
        "6": -12.712062,
        "7": -16.505384,
        "8": -19.821489,
        "9": -24.113360,
        "10": -28.422532,
        "11": -32.765970,
        "12": -37.967600,
        "13": -44.326801,
        "14": -47.845157,
        "15": -52.322627,
        "20": -77.177043,
        "25": -102.372663,
        "30": -128.286571,
        "38": -173.928427,
        "50": -244.549926,
        "100": -557.039820,
        "104": -582.038429,
    }

    benchmark = None

    def __init__(self, n_atoms: int = 10, dims=None, **kwargs):
        self.n_atoms = n_atoms
        self.dims = 3 * n_atoms
        self.benchmark = [judo.zeros(self.n_atoms * 3), self.minima.get(str(int(n_atoms)), 0)]
        super(LennardJones, self).__init__(dims=self.dims, function=lennard_jones, **kwargs)

    @staticmethod
    def get_bounds(dims):
        bounds = [(-15, 15) for _ in range(dims)]
        return Bounds.from_tuples(bounds)

    def __reset(self, **kwargs):
        states = super(LennardJones, self).reset()
        new_states = random_state.normal(0, scale=1.0, size=states.states.shape)
        states.update(observs=new_states, states=judo.copy(new_states))
        return states


ALL_BENCHMARKS = [Sphere, Rastrigin, EggHolder, StyblinskiTang, HolderTable, Easom]



"""This modes implements an interfacing with the `evotorch` library."""
from typing import Optional, Union

from evotorch.algorithms.searchalgorithm import SearchAlgorithm
from evotorch.core import Problem, SolutionBatch
import judo
from judo import Bounds, dtype

from fragile.core.env import Function
from fragile.core.typing import StateData


class EvotorchEnv(Function):
    """
    This environment implements an interface with the `evotorch` library.

    When providing an instance of an evotorch :class:`Searcher` it will \
    wrap it and allow to use all the fragile features on top of evotorch, such as \
    plotting, custom policies, etc.
    """

    default_inputs = {
        "actions": {},
        "observs": {"clone": True},
        "rewards": {"clone": True},
    }
    default_outputs = ("observs", "rewards", "oobs", "out_scores")

    def __init__(
        self,
        algorithm: SearchAlgorithm,
        **kwargs,
    ):
        """
        Initialize an :class:`EvotorchEnv` instance.

        Args:
            searcher (SearchAlgorithm): An instance of an evotorch :class:`SearchAlgorithm`.
            kwargs: Additional keyword arguments to pass to the :class:`Function`.
        """
        self._algorithm = algorithm
        reward_function = self._get_function()
        bounds = self._get_bounds()
        super().__init__(function=reward_function, bounds=bounds, **kwargs)

    @property
    def algorithm(self) -> SearchAlgorithm:
        """Return the evotorch :class:`SearchAlgorithm` instance used by this environment."""
        return self._algorithm

    @property
    def population(self) -> SolutionBatch:
        """
        Access the :class:`SolutionBatch` instance used by the evotorch :class:`SearchAlgorithm`.
        """
        return self._algorithm.population

    @property
    def problem(self) -> Problem:
        """Access the :class:`Problem` instance used by the evotorch :class:`SearchAlgorithm`."""
        return self._algorithm.problem

    @property
    def dtype(self):
        """
        Access the dtype used by the evotorch :class:`SearchAlgorithm` for the solution values.
        """
        return self._algorithm.population.dtype

    @property
    def eval_dtype(self):
        """Access the dtype used by the :class:`SearchAlgorithm` for the solution evaluations."""
        return self._algorithm.population.eval_dtype

    @property
    def solution_length(self) -> int:
        """Access the length of the solution used by the evotorch :class:`SearchAlgorithm`."""
        return self._algorithm.population.solution_length

    def _get_bounds(self) -> Bounds:
        """
        Initialize the :class:`Bounds` instance used by this environment.

        Extract all the information about the dimensionality and data type of the solutions
        from the problem instance used by the evotorch :class:`SearchAlgorithm`.
        """
        if self.problem.lower_bounds is not None:
            low = self.problem.lower_bounds
        else:
            low = self.problem.initial_lower_bounds
        if self.problem.upper_bounds is not None:
            high = self.problem.upper_bounds
        else:
            high = self.problem.initial_upper_bounds
        _dtype = self.dtype if judo.Backend.is_torch() else dtype.float32
        low = judo.to_backend(low)
        high = judo.to_backend(high)
        return Bounds(
            low=low,
            high=high,
            dtype=low.dtype if hasattr(low, "dtype") else _dtype,
            shape=self.solution_length,
        )

    def _get_function(self):
        """
        Return a function that sets the values of the evotorch :class:`SolutionBatch` \
        with the provided points and iterates the evotorch :class:`SearchAlgorithm` \
        to obtain new solutions.
        """

        def step_searcher(observs, rewards):
            self.algorithm.population.access_values()[:] = judo.to_torch(observs)
            self.algorithm.population.access_evals()[:] = judo.to_torch(rewards)
            self.algorithm.step()
            population = self.algorithm.population  # .clone()
            observs = judo.to_backend(population.access_values(keep_evals=True))
            rewards = judo.to_backend(population.access_evals())
            return observs, rewards

        return step_searcher

    def step(self, actions, observs, **kwargs):
        """
        Sum the target action to the observations to obtain the new points, and \
        evaluate the reward and boundary conditions.

        Returns:
            Dictionary containing the information of the new points evaluated.

             ``{"observs": new_points, "rewards": scalar array, \
             "oobs": boolean array}``

        """
        new_points_in = actions + observs if self._actions_as_perturbations else actions
        curr_reward = self.get("rewards").reshape(-1, 1)
        new_points, rewards = self.function(new_points_in, curr_reward)
        rewards = rewards.flatten()
        oobs = self.calculate_oobs(points=new_points, rewards=rewards)
        return dict(
            observs=new_points,
            rewards=rewards,
            oobs=oobs,
        )

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ) -> Union[None, StateData]:
        """
        Reset the :class:`Function` to the start of a new episode and updates its internal data.
        """
        if root_walker is None:
            new_points = self.sample_action(batch_size=self.swarm.n_actives)
            if self._x0 is not None:
                new_points[:] = self._x0
            elif self.start_same_pos:
                new_points[:] = new_points[0]
            actions = judo.zeros_like(new_points) if self._actions_as_perturbations else new_points
            curr_reward = self.get("rewards").reshape(-1, 1)
            new_points, rewards = self.function(new_points, curr_reward)
            rewards = rewards.flatten()
        else:
            new_points = self.get("observs")
            rewards = self.get("rewards")
            actions = (
                self.get("actions") if self._actions_as_perturbations else self.get("observs")
            )
        data = dict(
            actions=actions,
            observs=new_points,
            rewards=rewards,
        )
        if inplace:
            self.update(**data)
        else:
            return data
